! ******************************************************************
! ------------------------------------------------------------------
PROGRAM IMPORT_GRID
! ------------------------------------------------------------------
! Author: Astrid Kerkweg, Uni Mainz, July 2013 
!  based on NCREGRID by Patrick Joeckel, MPICH, Mainz, June 2002
! ******************************************************************

  USE mo_f2kcli                    ! command line interface
  USE messy_main_import_grid,     ONLY: READ_CONTROL, RG_CTRL,  RG_NML         &
                                      , RG_STATUS, RGSTAT_STOP, RG_PROC        &
                                      , RG_SCAN, NML_NEXT, NML_STAY
  USE messy_main_grid,            ONLY: t_geohybgrid, INIT_GEOHYBGRID          &
                                      , NEW_GEOHYBGRID, COPY_GEOHYBGRID
  USE messy_main_grid_netcdf,     ONLY: t_ncvar, QDEF_NCVAR, INIT_NCVAR        &
                                      , COPY_NCVAR, RGMSG, RGMLE
  USE messy_main_grid_trafo,      ONLY: GTRF_NONE, GTRF_NRGD, GTRF_SCRP        &
                                      , BALANCE_GEOHYBGRID, COMPLETE_GEOHYBGRID
  USE messy_main_grid_trafo_nrgd, ONLY: REGRID_CONTROL, ncregridvers            
  USE messy_main_grid_trafo_scrp, ONLY: t_scrip_data, SCRIP_CONTROL            &
                                      , CALC_SCRIP_WEIGHTS, CALC_SCRIPDATA     &
                                      , BALANCE_CURVILINEAR_PS                 &
                                      , construct_input_surf_pressure
  USE messy_main_tools,           ONLY: int2str, read_nml_open, read_nml_check &
                                      , read_nml_close

#ifdef PNCREGRID
  USE messy_main_import_grid_par, ONLY: INIT_PARALLEL
#endif

  IMPLICIT NONE

  INTRINSIC :: TRIM

#ifdef PNCREGRID
#include <mpif.h>
#endif
 
  ! FOR COMMAND LINE
  CHARACTER(LEN=256) :: EXE          ! program name
  CHARACTER(LEN=80)  :: CMD          ! argument
  INTEGER            :: NARG         ! number of arguments

  ! um_ak_20130725+
  LOGICAL,  SAVE                        :: lint    ! use input time ?
! op_pj_20140819+
!!$INTEGER,  POINTER                    :: RGT(:)        ! regridding type
  INTEGER, DIMENSION(:),  POINTER       :: RGT => NULL() ! regridding type
! op_pj_20140819-
  TYPE (t_ncvar), DIMENSION(:), POINTER :: rvar  => NULL() ! list of variables
  TYPE (t_ncvar), DIMENSION(:), POINTER :: ovar  => NULL() ! list of variables
  TYPE (t_ncvar), DIMENSION(:), POINTER :: sovar => NULL() ! list of variables
  TYPE (t_geohybgrid)                   :: igrid   ! output grid info
  TYPE (t_geohybgrid)                   :: ogrid   ! output grid info
  TYPE (t_geohybgrid)                   :: ogridtmp  ! output grid info
  TYPE (t_geohybgrid)                   :: intgrid ! intermediate grid info
  TYPE (t_geohybgrid)                   :: int2grid! intermediate grid info
  TYPE (t_geohybgrid)                   :: ipsgrid ! intermediate grid info
  TYPE (t_geohybgrid)                   :: intzogrid  ! output grid info
  INTEGER                               :: tt      ! time step
  INTEGER                               :: count = 0 ! procedure counter
  INTEGER                               :: i       ! counter
  TYPE(t_scrip_data), POINTER           :: PSD
  INTEGER                               :: igridid
  INTEGER                               :: ogridid
  CHARACTER(LEN=4)                      :: cstr = ''
  INTEGER                               :: status
  INTEGER                               :: SCRIP_ID     ! SCRIP DATA ID   
  LOGICAL                               :: lrgz         ! regrid in z direction
  INTEGER                               :: iipol
  ! um_ak_20130725-
  LOGICAL                               :: lex
  INTEGER                               :: fstat
  INTEGER                               :: tc, tmin, tmax, tret, tstep
  INTEGER,           PARAMETER          :: t_undef  = -99
  INTEGER                               :: ix
  LOGICAL                               :: lrgxy
  LOGICAL                               :: lpres
  LOGICAL                               :: lwork
  LOGICAL                               :: lpar

#ifdef PNCREGRID
  ! mz_kk_20080218+
  INTEGER         :: ierr, root
  INTEGER         :: rank, nproc
#endif

  NAMELIST /GTRF/ iipol

#ifdef PNCREGRID
  ! MPI Setup
  CALL MPI_Init(ierr)
  CALL MPI_Comm_rank(MPI_COMM_WORLD, rank,  ierr)
  CALL MPI_Comm_size(MPI_COMM_WORLD, nproc, ierr)
 
  CALL INIT_PARALLEL(rank, nproc, MPI_COMM_WORLD)
  lpar = .TRUE.
  write (*,*) 'SPECIES PARALLELSATION ON', nproc
#else
  lpar = .FALSE.
  write (*,*) 'SPECIES PARALLELSATION OFF'
#endif

  NARG = COMMAND_ARGUMENT_COUNT()    ! number of arguments
  CALL GET_COMMAND_ARGUMENT(0,EXE)   ! program name

  IF (NARG >= 2) THEN 
     WRITE(*,*) 'Too many arguments !'
     CALL USAGE(TRIM(EXE)) 
     STOP
  END IF

  IF (NARG == 0) THEN 
     CALL USAGE(TRIM(EXE)) 
     STOP
  END IF

  CALL GET_COMMAND_ARGUMENT(1,CMD)  

  ! SET DEFAULT
  iipol = 1

  INQUIRE(file=TRIM(CMD), exist=lex)
  IF (.not.lex)  THEN  ! CMD does not exist
     WRITE(*,*) '*** WARNING: FILE '''//TRIM(CMD)//'''  NOT FOUND !'
     STOP    
  END IF

  ! OPEN FILE
  OPEN(100,file=TRIM(CMD))
  WRITE(*,*) 'Reading namelist GTRF'
  REWIND(100)
  READ(100, NML=GTRF, IOSTAT=fstat)
  IF (fstat /= 0) THEN
     WRITE(*,*) '*** ERROR: READ ERROR in NAMELIST GTRF in FILE '''//TRIM(CMD)//''' !', fstat
  ELSE
     WRITE(*,*) ' ... OK !'
  END IF
  CLOSE(100)

  IF (iipol == GTRF_SCRP) THEN
     write (*,*) '*** INFO: SCRIP INTERPOLATION (iipol =',iipol,')'
  ELSE IF (iipol == GTRF_NRGD) THEN
     write (*,*) '*** INFO: NCREGRID INTERPOLATION (iipol =',iipol,')'
  ELSE IF (iipol == GTRF_NONE) THEN
     write (*,*) '*** INFO: NO INTERPOLATION (iipol =',iipol,')'
     STOP
  ELSE
     write (*,*) ' WRONG INTERPOLATION OPTION: ', iipol
     STOP
  ENDIF

  ! REGRIDDING ...
  RG_CTRL = RG_SCAN
  RG_NML  = NML_NEXT
  regrid_loop: DO ! endless DO loop (must be terminated with EXIT)
     ! um_ak_20130725+
     CALL INIT_GEOHYBGRID(igrid)
     CALL INIT_GEOHYBGRID(ogrid)
     CALL INIT_GEOHYBGRID(ogridtmp)
     igridid = -99
     ogridid = -99

     NULLIFY(rvar)

     tmin  = t_undef
     tmax  = t_undef
     tstep = t_undef
     tret  = t_undef
     ! um_ak_20130725-

     CALL READ_CONTROL(RG_CTRL, RG_NML, RG_STATUS      &
            , rvar, igrid, ogrid, RGT, lint, TRIM(CMD) &
            , tc=1, tmin=tmin, tmax=tmax, tstep=tstep  &
            , iounit=100, lvarparallel=lpar            &
            , lpresaxis=lpres, lwork=lwork )

     IF (RG_STATUS == RGSTAT_STOP) EXIT ! leave endless DO loop
!     REWIND(100)
     RG_NML = NML_STAY

     time_loop: DO tc=tmin,tmax,tstep
        CALL READ_CONTROL(RG_CTRL, RG_NML, RG_STATUS    &
             , rvar, igrid, ogrid, RGT, lint, TRIM(CMD) &
             , tc=tc ,  iounit=100, lvarparallel=lpar   )
        
        IF (RG_STATUS == RGSTAT_STOP) EXIT regrid_loop ! leave endless DO loop
        SELECT CASE(iipol)
        CASE(GTRF_NONE)
           ! NOT IMPLEMENTED FOR BOX MODEL
        CASE(GTRF_NRGD)
           write (0,*) 'lwork',lwork
           IF (lwork) &
           CALL REGRID_CONTROL(igrid, ogrid, rvar, ovar, RGT, lint &
                , lfirsto=(tc==tmin)) 
           
        CASE(GTRF_SCRP)
           CALL COMPLETE_GEOHYBGRID(igrid)
           CALL COMPLETE_GEOHYBGRID(ogrid)

! um_ak_20150714+
! generic grid names are now generated in new_geohybgrid
!!$           COUNT = COUNT + 1
!!$           CALL int2str(cstr, COUNT, '0')
!!$           igrid%name = 'ingrid'//cstr
! um_ak_20150714-
           CALL NEW_GEOHYBGRID(status, igridid, igrid) 
           IF (status /= 0 .AND. status /= 1) THEN
              CALL RGMSG('import_grid', RGMLE &
                   , 'NEW_GEOHYBGRID status ', status &
                   ,' for input grid', .TRUE.)
              EXIT
           END IF
           CALL NEW_GEOHYBGRID(status, ogridid, ogrid) 
           IF (status /= 0 .AND. status /= 1)  THEN
              CALL RGMSG('import_grid', RGMLE &
                   , 'NEW_GEOHYBGRID status ', status &
                   ,' for output grid', .TRUE.)
              EXIT
           END IF
           
           lrgxy = QDEF_NCVAR(ogrid%lonm)  .OR. QDEF_NCVAR(ogrid%loni)  .OR. &
                   QDEF_NCVAR(ogrid%clonm) .OR. QDEF_NCVAR(ogrid%cloni) .OR. &
                   QDEF_NCVAR(ogrid%rlonm) .OR. QDEF_NCVAR(ogrid%rloni) .OR. &
                   QDEF_NCVAR(ogrid%latm)  .OR. QDEF_NCVAR(ogrid%lati)  .OR. &
                   QDEF_NCVAR(ogrid%clatm) .OR. QDEF_NCVAR(ogrid%clati) .OR. &
                   QDEF_NCVAR(ogrid%rlatm) .OR. QDEF_NCVAR(ogrid%rlati)

           lrgz  = QDEF_NCVAR(ogrid%hybm)  .OR. QDEF_NCVAR(ogrid%hybi)  .OR. &
                   QDEF_NCVAR(ogrid%hyam)  .OR. QDEF_NCVAR(ogrid%hyai)

           write (*,*) ' Horizontal / Vertical Interpolation ?', lrgxy, lrgz
              
           IF (.NOT. lrgxy .AND. .NOT. lrgz) THEN
              write (*,*) ' !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
              write (*,*) '   NEITHER horizontal                        '
              write (*,*) '      NOR  vertical interpolation            ' 
              write (*,*) '              requested:                     '
              write (*,*) '                CYCLE                        '
              write (*,*) ' !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
              EXIT
           ENDIF

           CALL COPY_GEOHYBGRID(ogridtmp, ogrid)
           IF (lrgz .AND. .NOT. lrgxy) THEN
              CALL BALANCE_GEOHYBGRID(igrid, ogridtmp)
           ENDIF

           IF (lrgxy .OR. lrgz) THEN
              CALL CALC_SCRIPDATA(status, igrid, ogridtmp, RGT, SCRIP_ID &
                   , PSD=PSD)
              IF (status /= 0 ) THEN
                 IF (status /= 01 )  THEN ! status == 01 : PSD exists already
                    CALL RGMSG('import_grid', RGMLE, 'calc_scrip_data ', status &
                         ,' ', .TRUE.)
                    EXIT
                 ENDIF
              ELSE
                 ! CALCULATE WEIGHTS
                 CALL CALC_SCRIP_WEIGHTS(status, PSD)
                 IF (status /= 0) THEN
                    CALL RGMSG('import_grid', RGMLE, 'calc_scrip_weights ' &
                         , status ,' ', .TRUE.)
                    EXIT
                 END IF
              END IF
           ENDIF

           IF (lrgxy) THEN
              CALL SCRIP_CONTROL(status, SCRIP_ID, igrid, ogrid, RGT, lint&
                   , rvar, sovar,intgrid, llrgz=lrgz, lfirsto=(tc==tmin))
              IF (status /= 0) THEN
                 CALL RGMSG('import_grid', RGMLE, 'SCRIP_CONTROL ', status &
                      ,' ', .TRUE.)
                 EXIT
              END IF
           ELSE
              ALLOCATE(sovar(SIZE(rvar)))
              DO ix = 1, SIZE(rvar)
                 CALL COPY_NCVAR(sovar(ix), rvar(ix))
              END DO
              CALL COPY_GEOHYBGRID(intgrid, igrid)
           ENDIF

           ! vertical interpolation
           IF (lrgz) THEN
              
              ! make grid with vertical input grid, but horizontal
              ! SCRIP-gridded grid 
              CALL COPY_GEOHYBGRID(int2grid,  igrid)
              CALL COPY_GEOHYBGRID(intzogrid, ogrid)
              
              IF ((QDEF_NCVAR(ogrid%rlonm) .AND. QDEF_NCVAR(ogrid%rlatm)) &
                   .OR. &
                   (QDEF_NCVAR(ogrid%rloni) .AND. QDEF_NCVAR(ogrid%rlati)))&
                   THEN
                 
                 ! CONSTRUCT CORRECT PS-INPUT-GRID FOR VERTICAL
                 ! INTERPOLATION

                 ! SURFACE PRESSURE needs to be available for both grid
                 ! on the rotated grid
                 CALL COPY_GEOHYBGRID(ipsgrid, igrid)

                 ! assume congruency of grids ogrid and intgrid
                 ! take rotated coordinates for curvi-linear grids
                 ! instead of intermediate grid. NCREGRID can only 
                 ! handle lat-lon-grids
                 ipsgrid%lonc = .FALSE.
                 CALL COPY_NCVAR(ipsgrid%lonm, ogrid%rlonm)
                 CALL COPY_NCVAR(ipsgrid%loni, ogrid%rloni)
                 CALL COPY_NCVAR(ipsgrid%latm, ogrid%rlatm)
                 CALL COPY_NCVAR(ipsgrid%lati, ogrid%rlati)
                 
                 CALL INIT_NCVAR(ipsgrid%clonm)
                 CALL INIT_NCVAR(ipsgrid%cloni)
                 CALL INIT_NCVAR(ipsgrid%clatm)
                 CALL INIT_NCVAR(ipsgrid%clati)
                 CALL INIT_NCVAR(ipsgrid%rlonm)
                 CALL INIT_NCVAR(ipsgrid%rloni)
                 CALL INIT_NCVAR(ipsgrid%rlatm)
                 CALL INIT_NCVAR(ipsgrid%rlati)
 
                 CALL INIT_NCVAR(ipsgrid%ps)
                 CALL CONSTRUCT_INPUT_SURF_PRESSURE(status, igrid, ipsgrid &
                                                          , SCRIP_ID)
                 IF (status /= 0) THEN
                    CALL RGMSG('import_grid', RGMLE, 'CONSTRUCT_PS ' &
                         , status,' ', .TRUE.)
                    EXIT
                 ENDIF
                 CALL BALANCE_CURVILINEAR_PS(status, ipsgrid, ogrid &
                      , SCRIP_ID)
                 IF (status /= 0) THEN
                    CALL RGMSG('import_grid', RGMLE, 'BALANCE_CURVILINEAR_PS ' &
                         , status,' ', .TRUE.)
                    EXIT
                 ENDIF

                 ! replace in int2grid for vertical interpolation
                 CALL COPY_GEOHYBGRID(int2grid, ipsgrid)
                 CALL INIT_GEOHYBGRID(ipsgrid)

                 ! CONSTRUCT OUTPUT GRID FOR VERTICAL INTERPOLATION
                 intzogrid%lonc = .FALSE.
                 CALL COPY_NCVAR(intzogrid%lonm, ogrid%rlonm)
                 CALL COPY_NCVAR(intzogrid%loni, ogrid%rloni)
                 CALL COPY_NCVAR(intzogrid%latm, ogrid%rlatm)
                 CALL COPY_NCVAR(intzogrid%lati, ogrid%rlati)
                 
                 CALL INIT_NCVAR(intzogrid%clonm)
                 CALL INIT_NCVAR(intzogrid%cloni)
                 CALL INIT_NCVAR(intzogrid%clatm)
                 CALL INIT_NCVAR(intzogrid%clati)
                 CALL INIT_NCVAR(intzogrid%rlonm)
                 CALL INIT_NCVAR(intzogrid%rloni)
                 CALL INIT_NCVAR(intzogrid%rlatm)
                 CALL INIT_NCVAR(intzogrid%rlati)
              ELSE
                 ! no curvi-linear grid, used horizontal grid from
                 ! intermediate grid
                 CALL COPY_NCVAR(int2grid%lonm, intgrid%lonm)
                 CALL COPY_NCVAR(int2grid%loni, intgrid%loni)
                 CALL COPY_NCVAR(int2grid%latm, intgrid%latm)
                 CALL COPY_NCVAR(int2grid%lati, intgrid%lati)

                 CALL CONSTRUCT_INPUT_SURF_PRESSURE(status, igrid, int2grid &
                                                          , SCRIP_ID)
                 IF (status /= 0) THEN
                    CALL RGMSG('import_grid', RGMLE, 'CONSTRUCT PS ' &
                         , status,' ', .TRUE.)
                    EXIT
                 ENDIF

                 CALL BALANCE_CURVILINEAR_PS(status, int2grid, ogrid &
                      , SCRIP_ID)
                 IF (status /= 0) THEN
                    CALL RGMSG('import_grid', RGMLE, 'BALANCE_CURVILINEAR_PS ' &
                         , status,' ', .TRUE.)
                    EXIT
                 ENDIF
             ENDIF
              
             IF (lwork) &
                  CALL REGRID_CONTROL(int2grid, intzogrid, sovar, ovar, RGT  &
                   , lint, lrgx=.FALSE., lrgy=.FALSE., lrgz=lrgz             &
                   , lfirsto=(tc==tmin)) 
              
              CALL INIT_GEOHYBGRID(int2grid)
              CALL INIT_GEOHYBGRID(intzogrid)
           ENDIF
        END SELECT

        CALL INIT_GEOHYBGRID(intgrid)
        
        IF (ASSOCIATED(rvar)) THEN
           DO i=1,SIZE(rvar)
              CALL INIT_NCVAR(rvar(i))
           END DO
           DEALLOCATE(rvar)
           NULLIFY(rvar)
        ENDIF

        IF (ASSOCIATED(ovar)) THEN
           DO i=1, SIZE(ovar)
              CALL INIT_NCVAR(ovar(i))
           END DO
           DEALLOCATE(ovar)
           NULLIFY(ovar) 
        ENDIF
        
        IF (ASSOCIATED(sovar)) THEN
           DO i=1, SIZE(sovar)
              CALL INIT_NCVAR(sovar(i))
           END DO
           DEALLOCATE(sovar)
           NULLIFY(sovar) 
        ENDIF
        
     END DO time_loop
     RG_CTRL = RG_PROC
     RG_NML  = NML_NEXT
        
  END DO regrid_loop
  ! END OF REGRIDDING

#ifdef PNCREGRID
  CALL MPI_Finalize(ierr)  ! mz_kk_20080218
#endif

CONTAINS

  SUBROUTINE USAGE(EXE)
    CHARACTER (LEN=*) :: EXE
    WRITE(*,*) '--------------------------------------------'
    WRITE(*,*) 'IMPORT_GRID Version 0.9b'
    WRITE(*,*) 'Author: Astrid Kerkweg, Uni Mainz, July 2013'
    WRITE(*,*) '--------------------------------------------'
    WRITE(*,*) 'based on NCREGRID Version ',NCREGRIDVERS
    WRITE(*,*) 'Author: Patrick Joeckel, MPICH, June 2002'
#ifdef PNCREGRID
    WRITE(*,*) '--------------------------------------------'
    WRITE(*,*) 'MPI parallelisation (number of variables)'
!!$    WRITE(*,*) 'Author: Klaus Ketelsen, MPICH, Dec 2007'
#endif
    WRITE(*,*) '--------------------------------------------'
    WRITE(*,*) 'Usage: '//TRIM(EXE)//' <namelist-file>'
    WRITE(*,*) '--------------------------------------------'
  END SUBROUTINE USAGE

END PROGRAM IMPORT_GRID
! ------------------------------------------------------------------
