! ******************************************************************
! ------------------------------------------------------------------
MODULE MESSY_MAIN_IMPORT_GRID
! ------------------------------------------------------------------
! Author: Patrick Joeckel, MPICH, Mainz, October 2002
!         extended for curvilinear ocean grids by
!         Astrid  Kerkweg, UniMz, Mainz, 2012-2013
!         -> re-structured/expanded for more general GRID application
! ******************************************************************

  USE MESSY_MAIN_CONSTANTS_MEM, ONLY: DP, STRLEN_MEDIUM
  USE MESSY_MAIN_IMPORT_GRID_PAR

  IMPLICIT NONE
  PRIVATE

  PUBLIC :: DP

  INTRINSIC :: ABS, ACHAR, ALLOCATED, ASSOCIATED, DATE_AND_TIME, INDEX &
             , LEN_TRIM, MIN, NULL, PRESENT, REAL, SIZE, TINY, TRIM &
             , ADJUSTL

  ! GLOBAL PARAMETERS
  CHARACTER(LEN=*), PARAMETER, PUBLIC :: submodstr = 'import_grid'
  CHARACTER(LEN=*), PARAMETER, PUBLIC :: submodver = '1.0'

  INTEGER,           PARAMETER   :: NCCNTMAXNLEN = 40                  !   ! 
  CHARACTER(LEN=24), PARAMETER   :: fstr = '(a80,1x,a40,1x,4(i8,1x))'  ! <-V
  INTEGER,           PARAMETER   :: NCCRSTRL = 80                      !   |

  ! TYPE DECLARATION TO HOLD COUNTER INFORMATION
  TYPE t_ncrgcnt
     CHARACTER(LEN=NCCNTMAXNLEN) :: name = ''
     INTEGER                     :: start = 1
     INTEGER                     :: step  = 1
     INTEGER                     :: reset = 1
     INTEGER                     :: current = 1
  END TYPE t_ncrgcnt
  
  ! LIST TO STORE COUNTER INFORMATION FOR RESTART
  TYPE t_ncrgcnt_list
     CHARACTER(NCCRSTRL)           :: mname  = ''
     TYPE(T_NCRGCNT)               :: this
     TYPE(T_NCRGCNT_LIST), POINTER :: next => NULL()
  END TYPE t_ncrgcnt_list

  ! ====================================================================
  ! CONCAT. LIST OF RGT-EVENTS
  TYPE(t_ncrgcnt_list), POINTER, SAVE :: GRGTLIST => NULL()
  ! ====================================================================
  
  INTEGER,  PUBLIC, SAVE      :: RG_CTRL, RG_NML, RG_STATUS
  INTEGER,  PUBLIC, PARAMETER :: RG_SCAN = 1
  INTEGER,  PUBLIC, PARAMETER :: RG_PROC = 2
  INTEGER,  PUBLIC, PARAMETER :: RG_STOP = 3
  INTEGER,  PUBLIC, PARAMETER :: NML_NEXT = 4
  INTEGER,  PUBLIC, PARAMETER :: NML_STAY = 5
  INTEGER,  PUBLIC, PARAMETER :: RGSTAT_START = 6
  INTEGER,  PUBLIC, PARAMETER :: RGSTAT_CONT = 7
  INTEGER,  PUBLIC, PARAMETER :: RGSTAT_STOP = 8

  INTEGER, SAVE :: COUNT = 0

  PUBLIC :: parse_str

  ! COUNTER INFORMATION HANDLING
  PUBLIC  :: t_ncrgcnt          ! TYPE STRUCT FOR COUNTER INFORMATION I/O
  PUBLIC  :: RGTOOL_NCRGCNT_RST ! MANAGES COUNTER INFORMATION RESTART I/O
  !
  PUBLIC  :: CLEAN_NCRGCNT_LIST
  PUBLIC  :: WRITE_NCRGCNT_LIST
  PUBLIC  :: READ_NCRGCNT_LIST
  PUBLIC  :: GET_NEXT_NCRGCNT
  !
  !PRIVATE :: LOC_NCRGCNT
  !PRIVATE :: NEW_NCRGCNT
  !PRIVATE :: NCRGCNT_ERROR_STR
  PUBLIC :: PARSE_VARSTR
  PUBLIC :: READ_NAMELIST
  PUBLIC :: READ_CONTROL

  PUBLIC  :: RGTOOL_READ_NCVAR  ! NCREGRID ONE STEP OF ONE FIELD
  PUBLIC  :: RGTOOL_READ_NCFILE ! NCREGRID ONE STEP OF ONE FILE

CONTAINS

! ----------------------------------------------------------------------
  SUBROUTINE parse_str(status, strlen, str, nml, var, file, z, heights, ixf &
       , ipol,grdname)

    USE MESSY_MAIN_TOOLS,      ONLY: STRCRACK
    USE MESSY_MAIN_GRID_TRAFO, ONLY: GTRFCHAR, GTRF_MAX

    IMPLICIT NONE

    INTRINSIC :: TRIM, ADJUSTL

    ! I/O
    INTEGER,          INTENT(OUT)   :: status   ! status information
    INTEGER,          INTENT(IN)    :: strlen   ! max. length of strings
    CHARACTER(LEN=*), INTENT(IN)    :: str      ! string to parse
    CHARACTER(LEN=*), INTENT(INOUT) :: nml      ! namelist file
    CHARACTER(LEN=*), INTENT(INOUT) :: var      ! netCDF variable
    CHARACTER(LEN=*), INTENT(INOUT) :: file     ! netCDF file
    REAL(DP), DIMENSION(:), POINTER :: z        ! level information
    CHARACTER(LEN=*), INTENT(INOUT) :: heights  ! string containing heights
    INTEGER,          INTENT(INOUT) :: ixf      ! index regridding
    INTEGER         , INTENT(INOUT) :: ipol     ! interpolation method:
                                                ! NRGD : NREGRID
                                                ! SCRP : SCRIP
                                                ! NONE : no regridding
    CHARACTER(LEN=STRLEN_MEDIUM), INTENT(INOUT) :: grdname

    ! LOCAL
    CHARACTER(LEN=strlen), POINTER :: sl1(:)
    CHARACTER(LEN=strlen), POINTER :: sl2(:)
    CHARACTER(LEN=strlen), POINTER :: sl3(:)
    INTEGER                        :: n, m, l
    INTEGER                        :: i, j
    INTEGER                        :: iostat

    status = 1 ! ERROR

    NULLIFY(sl1)
    NULLIFY(sl2)
    NULLIFY(sl3)

    IF (ASSOCIATED(z)) THEN
       DEALLOCATE(z)
       NULLIFY(z)
    END IF

    CALL STRCRACK(str, ';', sl1, n)
    DO i=1, n

       CALL STRCRACK(sl1(i), '=', sl2, m)
       IF (SIZE(sl2) == 2) THEN
          IF (TRIM(ADJUSTL(sl2(2))) == '') THEN
             status = 2    ! EMPTY SPECIFICATION
             RETURN
          END IF
       END IF

       SELECT CASE(TRIM(ADJUSTL(sl2(1))))
          CASE('NML')
             nml = TRIM(ADJUSTL(sl2(2)))
          CASE('VAR')
             var = TRIM(ADJUSTL(sl2(2)))
          CASE('FILE')
             file = TRIM(ADJUSTL(sl2(2)))
          CASE('Z')
             heights=TRIM(ADJUSTL(sl2(2)))
             CALL STRCRACK(TRIM(ADJUSTL(sl2(2))), ',', sl3, l)
             ALLOCATE(z(l))
             DO j=1, l
                READ(sl3(j),*,IOSTAT=iostat) z(j)
                IF (iostat /= 0) THEN
                   status = 6  ! ERROR IN READING REAL
                   RETURN
                END IF
             END DO
          CASE('IPOL')
             DO j=1,GTRF_MAX
                IF (TRIM(ADJUSTL(sl2(2))) == GTRFCHAR(j)) THEN
                   ipol = j
                   EXIT
                ENDIF
             END DO
          CASE ('GRID')
             grdname=TRIM(ADJUSTL(sl2(2)))
          CASE('IXF')
             READ(sl2(2),*,IOSTAT=iostat) ixf
                IF (iostat /= 0) THEN
                   status = 7  ! ERROR IN READING INTEGER
                   RETURN
                END IF
          CASE DEFAULT
             status = 3 ! UNKNOWN SPECIFIER
             RETURN
       END SELECT

    END DO

    ! CLEAN UP
    IF (ASSOCIATED(sl1)) DEALLOCATE(sl1)
    IF (ASSOCIATED(sl2)) DEALLOCATE(sl2)

    status = 0 ! NO ERROR

  END SUBROUTINE parse_str
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE CLEAN_NCRGCNT_LIST(status)

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED

    ! I/O
    INTEGER,                INTENT(OUT) :: status

    ! LOCAL
    TYPE(t_ncrgcnt_list), POINTER :: ai => NULL()
    TYPE(t_ncrgcnt_list), POINTER :: ae => NULL()

    ! INIT
    status = 1
    
    ai => GRGTLIST
    DO
       IF (.NOT. ASSOCIATED(ai)) EXIT
       !
       ae => ai
       ai => ae%next
       !
       DEALLOCATE(ae)
       NULLIFY(ae)
       !
    END DO    

    NULLIFY(GRGTLIST)

    status = 0

  END SUBROUTINE CLEAN_NCRGCNT_LIST
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE WRITE_NCRGCNT_LIST(status, fname, iou)

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED, TRIM

    ! I/O
    INTEGER,          INTENT(OUT) :: status
    CHARACTER(LEN=*), INTENT(IN)  :: fname
    INTEGER,          INTENT(IN)  :: iou

    ! LOCAL
    CHARACTER(LEN=*),   PARAMETER :: substr = 'WRITE_NCRGCNT_LIST'
    TYPE(t_ncrgcnt_list), POINTER :: ai => NULL()
    TYPE(t_ncrgcnt_list), POINTER :: ae => NULL()

    ! EMPTY LIST
    IF (.NOT. ASSOCIATED (GRGTLIST)) THEN
       status = 0
       RETURN
    END IF

    WRITE(*,*) substr//&
         &': WRITING COUNTER LIST TO FILE '''//TRIM(fname)//''' ...'

    OPEN(iou, file=TRIM(fname), status='UNKNOWN')
    
    ai => GRGTLIST
    DO
       IF (.NOT. ASSOCIATED(ai)) EXIT
       !
       ae => ai
       ai => ae%next
       !
       WRITE(*,*) '--> ',TRIM(ae%mname)//' @ '//TRIM(ae%this%name)
       WRITE(*,*) '    ', ae%this%start, &
            ae%this%step, ae%this%reset, ae%this%current
       !
       WRITE(iou, fstr) ae%mname, ae%this%name, ae%this%start, &
            ae%this%step, ae%this%reset, ae%this%current
       !
    END DO    

    CLOSE(iou)

    WRITE(*,*) substr//&
         &': ... END OF COUNTER LIST REACHED !'

    status = 0

  END SUBROUTINE WRITE_NCRGCNT_LIST
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE READ_NCRGCNT_LIST(status, fname, iou)

    USE MESSY_MAIN_GRID_NETCDF, ONLY: RGMSG, RGMLE, RGMLEC, RGMLI, RGMLIC

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED, TRIM

    ! I/O
    INTEGER,          INTENT(OUT) :: status
    CHARACTER(LEN=*), INTENT(IN)  :: fname
    INTEGER,          INTENT(IN)  :: iou

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'READ_NCRGCNT_LIST'
    LOGICAL                  :: lex
    INTEGER                  :: fstat
    TYPE(t_ncrgcnt), POINTER :: cntptr => NULL()
    CHARACTER(LEN=NCCRSTRL)  :: mname
    TYPE(t_ncrgcnt)          :: ch
    
    ! EMPTY LIST
    IF (.NOT. ASSOCIATED (GRGTLIST)) THEN
       status = 0
       RETURN
    END IF

    CALL RGMSG(substr, RGMLI, &
         'OPENING COUNTER FILE '''//TRIM(fname)//''' ...')

    INQUIRE(file=TRIM(fname), exist=lex)
    IF (.NOT.lex) THEN  ! ERROR
       CALL RGMSG(substr, RGMLE, &
            'COUNTER FILE NOT FOUND: '//TRIM(fname) )
       status = 5020 ! COUNTER FILE NOT FOUND
       RETURN
    END IF

    OPEN(iou, file=TRIM(fname), status='OLD')

    DO
       READ(iou,fstr, IOSTAT=fstat) &
            mname, ch%name,ch%start,ch%step,ch%reset,ch%current
       IF (fstat /= 0) EXIT
       !
       CALL RGMSG(substr, RGMLIC, &
            '... LOOKING FOR COUNTER '''//TRIM(mname)//' @ '&
            &//TRIM(ch%name)//''' ...')
       !
       CALL LOC_NCRGCNT(status, TRIM(mname), TRIM(ch%name), cntptr)
       IF ((status /= 0) .AND. (status /= 5005)) THEN
          CALL NCRGCNT_HALT(substr, status)
       END IF
       !
       IF (status == 0) THEN
          CALL RGMSG(substr, RGMLIC, '... UPDATING COUNTER:')
          ! UPDATE
          cntptr%start   = ch%start
          cntptr%step    = ch%step
          cntptr%reset   = ch%reset
          cntptr%current = ch%current
          CALL RGMSG(substr, RGMLIC, &
               '         START  : ',cntptr%start,' ')
          CALL RGMSG(substr, RGMLIC, &
               '         STEP   : ',cntptr%step,' ')
          CALL RGMSG(substr, RGMLIC, &
               '         RESET  : ',cntptr%reset,' ')
          CALL RGMSG(substr, RGMLIC, &
               '         CURRENT: ',cntptr%current,' ')
       ELSE
          ! COUNTER DOES NOT EXIST
          CALL RGMSG(substr, RGMLIC, '... COUNTER NOT FOUND')
       END IF
       !
    END DO

    IF (fstat < 0) THEN
       ! OK: END OF FILE REACHED
       CALL RGMSG(substr, RGMLIC, &
            'END OF FILE '''//TRIM(fname)//''' REACHED !')
    ELSE
       ! ERROR
       CLOSE(iou)
       CALL RGMSG(substr, RGMLE, &
            'READ ERROR READING FROM FILE '''//&
            &TRIM(fname)//''' !' , .false.)
       CALL RGMSG(substr, RGMLEC, &
            'STATUS: ',fstat,' ')
    END IF

    CLOSE(iou)

    status = 0

  END SUBROUTINE READ_NCRGCNT_LIST
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE RGTOOL_NCRGCNT_RST(mname, start, restart, event, c, lout, linit)

    ! MANAGES I/O OF COUNTER INFORMATION AT START AND RESTART
    !
    ! Author: Patrick Joeckel, MPICH, September 2005

    USE MESSY_MAIN_GRID_NETCDF, ONLY: RGMSG, RGMLI, RGMLIC

    IMPLICIT NONE

    INTRINSIC :: PRESENT, TRIM

    ! I/O
    CHARACTER(LEN=*), INTENT(IN)    :: mname
    LOGICAL,          INTENT(IN)    :: start   ! .true. at first time step
    LOGICAL,          INTENT(IN)    :: restart ! .true. at first time step of
                                               ! rerun
    LOGICAL,          INTENT(IN)    :: event   ! .true. on event
    TYPE(t_ncrgcnt),  INTENT(INOUT) :: c       ! counter - struct
    LOGICAL,          INTENT(IN)           :: lout
    LOGICAL,          INTENT(IN), OPTIONAL :: linit

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'RGTOOL_NCRGCNT_RST'
    TYPE(t_ncrgcnt),  POINTER   :: cntptr
    LOGICAL                     :: lf
    INTEGER                     :: status
    LOGICAL                     :: zlinit

    ! DO NOTHING IN CASE ...
    IF (.NOT.(start.OR.restart.OR.event)) THEN
       IF (lout) CALL RGMSG(substr, RGMLI, &
            'RGTEVENT '''//TRIM(mname)//' @ '//TRIM(c%name)//''' NOT ACTIVE !')
       RETURN
    ELSE
       IF (lout) CALL RGMSG(substr, RGMLI, &
            'RGTEVENT '''//TRIM(mname)//' @ '//TRIM(c%name)//''' ACTIVE !')
    END IF

    ! INIT
    IF (PRESENT(linit)) THEN
       zlinit = linit
    ELSE
       zlinit = .FALSE. ! DEFAULT
    END IF

    ! LOCATE COUNTER IN LIST
    IF (lout) CALL RGMSG(substr, RGMLIC, &
         '... LOOKING FOR COUNTER '''//TRIM(mname)//' @ '//&
         &TRIM(c%name)//''' ...')
    CALL LOC_NCRGCNT(status, TRIM(mname), TRIM(c%name), cntptr)
    IF ((status /= 0) .AND. (status /= 5005)) THEN
       IF (lout) CALL NCRGCNT_HALT(substr, status)
    END IF
    lf = (status == 0) ! FOUND ?
    IF (lf) THEN
       ! UPDATE FROM LIST
       IF (lout) CALL RGMSG(substr, RGMLIC, '... COPYING COUNTER FROM LIST:')
       c%start   = cntptr%start
       c%step    = cntptr%step
       c%reset   = cntptr%reset
       c%current = cntptr%current
       IF (lout) THEN
          CALL RGMSG(substr, RGMLIC, &
               '         START  : ',c%start,' ')
          CALL RGMSG(substr, RGMLIC, &
               '         STEP   : ',c%step,' ')
          CALL RGMSG(substr, RGMLIC, &
               '         RESET  : ',c%reset,' ')
          CALL RGMSG(substr, RGMLIC, &
               '         CURRENT: ',c%current,' ')
       END IF
    END IF

    ! START
    IF (zlinit) THEN
       IF (lf) THEN
          IF (lout) CALL NCRGCNT_HALT(substr, 5004) ! COUNTER EXISTS
       ELSE
          IF (lout) CALL RGMSG(substr, RGMLIC, &
               '... SAVING COUNTER '''//TRIM(mname)//' @ '//TRIM(c%name)&
               &//''' IN LIST (INIT)')
          ! SAVE IN LIST
          CALL NEW_NCRGCNT(status, TRIM(mname), c)
          IF (lout) CALL NCRGCNT_HALT(substr, status)
          ! RE-SET POINTER
          CALL LOC_NCRGCNT(status, TRIM(mname), TRIM(c%name), cntptr)
          IF (lout) CALL NCRGCNT_HALT(substr, status)
          lf = .TRUE. ! NOW IT EXISTS
          !
       END IF
    END IF

    ! EVENT
    IF (event.AND.(.NOT.(start))) THEN
       IF (.NOT. lf) THEN
          IF (lout) CALL NCRGCNT_HALT(substr, 5005) ! COUNTER DOES NOT EXIST
       ELSE
          IF (lout) CALL RGMSG(substr, RGMLIC, &
               '... UPDATING COUNTER '''//TRIM(mname)//' @ '//TRIM(c%name)&
               &//''' IN LIST (EVENT)')
          !
          ! UPDATE
          c%current = c%current + c%step                   ! INCREMENT
          IF (c%current > c%reset) c%current = c%start     ! RESET
          !
          ! SAVE IN LIST
          cntptr%start   = c%start
          cntptr%step    = c%step
          cntptr%reset   = c%reset
          cntptr%current = c%current
          !
          IF (lout)  THEN
             CALL RGMSG(substr, RGMLIC, &
                  '         START  : ',cntptr%start,' ')
             CALL RGMSG(substr, RGMLIC, &
                  '         STEP   : ',cntptr%step,' ')
             CALL RGMSG(substr, RGMLIC, &
                  '         RESET  : ',cntptr%reset,' ')
             CALL RGMSG(substr, RGMLIC, &
                  '         CURRENT: ',cntptr%current,' ')
          END IF
       END IF
    END IF

  END SUBROUTINE RGTOOL_NCRGCNT_RST
! ----------------------------------------------------------------------

  ! -------------------------------------------------------------------
  SUBROUTINE GET_NEXT_NCRGCNT(last, cntptr)

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED

    ! I/O
    LOGICAL,              INTENT(OUT) :: last
    TYPE(t_ncrgcnt),      POINTER     :: cntptr

    ! LOCAL
    TYPE(t_ncrgcnt_list), POINTER, SAVE :: ai  => NULL()
    TYPE(t_ncrgcnt_list), POINTER, SAVE :: ae  => NULL()
    INTEGER, PARAMETER                  :: MODE_INIT = 1
    INTEGER, PARAMETER                  :: MODE_CONT = 2
    INTEGER,                       SAVE :: MODE = MODE_INIT

    ! INIT
    last = .FALSE.

    DO
       SELECT CASE (MODE)
       CASE(MODE_INIT)
          ai => GRGTLIST
          MODE = MODE_CONT
       CASE(MODE_CONT)
          IF (.NOT. ASSOCIATED(ai)) EXIT
          ae => ai
          ai => ai%next
          cntptr => ae%this
          RETURN
       END SELECT
    END DO

    MODE = MODE_INIT
    last = .TRUE.

  END SUBROUTINE GET_NEXT_NCRGCNT
  ! -------------------------------------------------------------------

! ----------------------------------------------------------------------
! PRIVATE SUBROUTINES
! ----------------------------------------------------------------------

  ! -------------------------------------------------------------------
  SUBROUTINE loc_ncrgcnt(status, mname, cname, cntptr)

    IMPLICIT NONE

    INTRINSIC :: ADJUSTL, LEN_TRIM, TRIM, ASSOCIATED

    ! I/O
    INTEGER,              INTENT(OUT) :: status
    CHARACTER(LEN=*),     INTENT(IN)  :: mname
    CHARACTER(LEN=*),     INTENT(IN)  :: cname
    TYPE(t_ncrgcnt),      POINTER     :: cntptr

    ! LOCAL
    TYPE(t_ncrgcnt_list), POINTER :: ai  => NULL()
    TYPE(t_ncrgcnt_list), POINTER :: ae  => NULL()
    LOGICAL                       :: lex

    ! INIT
    lex = .FALSE.
    NULLIFY(cntptr)

    ! CHECKS
    IF (LEN_TRIM(ADJUSTL(mname)) > NCCRSTRL) THEN
       status = 5003   ! NAME TOO LONG
       RETURN
    END IF
    !
    IF (LEN_TRIM(ADJUSTL(cname)) > NCCRSTRL) THEN
       status = 5006   ! NAME TOO LONG
       RETURN
    END IF

    ! CHECK, IF IT EXISTS
    ai => GRGTLIST
    DO
       IF (.NOT. ASSOCIATED(ai)) EXIT
       IF ( (TRIM(ADJUSTL(mname)) == TRIM(ai%mname)) .AND. &
            (TRIM(ADJUSTL(cname)) == TRIM(ai%this%name)) ) THEN
          lex = .TRUE.
          EXIT
       END IF
       ae => ai
       ai => ae%next
    END DO    

    IF (lex) THEN
       cntptr => ai%this
    ELSE
       status = 5005   ! DOES NOT EXIST
       RETURN
    END IF

    status = 0

  END SUBROUTINE loc_ncrgcnt
  ! -------------------------------------------------------------------

  ! ----------------------------------------------------------------------
  SUBROUTINE new_ncrgcnt(status, mname, cnt)

    IMPLICIT NONE

    INTRINSIC :: ADJUSTL, LEN_TRIM, ASSOCIATED, TRIM, NULL

    ! I/O
    INTEGER,          INTENT(OUT)  :: status
    CHARACTER(LEN=*), INTENT(IN)   :: mname
    TYPE(t_ncrgcnt),  INTENT(IN)   :: cnt

    ! LOCAL
    TYPE(t_ncrgcnt_list), POINTER :: ai     => NULL()
    TYPE(t_ncrgcnt_list), POINTER :: ae     => NULL()
    TYPE(t_ncrgcnt),      POINTER :: cntptr => NULL()
    INTEGER                       :: zstat

    IF (LEN_TRIM(ADJUSTL(mname)) > NCCRSTRL) THEN
       status = 5003   ! SUBMODEL NAME TOO LONG
       RETURN
    END IF

    CALL loc_ncrgcnt(zstat, TRIM(mname), TRIM(cnt%name), cntptr)
    IF (zstat /= 5005) THEN  ! COUNTER DOES NOT EXIST (IS OK HERE !)
       IF (zstat == 0) THEN  ! COUNTER EXISTS ALREADY
          status = 5004      ! COUNTER EXISTS ALREADY
       ELSE
          status = zstat     ! ERROR
       END IF
       RETURN
    END IF

    ! GOTO END OF LIST
    ai => GRGTLIST
    DO
       IF (.NOT. ASSOCIATED(ai)) EXIT
       ae => ai
       ai => ai%next
    END DO

    ! ADD NEW
    ALLOCATE(ai)
    NULLIFY(ai%next)
    IF (.NOT. ASSOCIATED(GRGTLIST)) THEN
       GRGTLIST => ai                 ! SET POINTER TO FIRST ELEMENT
    ELSE
       ae%next => ai                   ! SET NEXT POINTER OF LAST ELEMENT
       !                               ! TO NEW ELEMENT
    END IF
       
    ! SET VALUES
    ai%mname        = TRIM(ADJUSTL(mname))
    !
    ai%this%name    = TRIM(ADJUSTL(cnt%name))
    ai%this%start   = cnt%start
    ai%this%step    = cnt%step
    ai%this%reset   = cnt%reset
    ai%this%current = cnt%current

    status = 0

  END SUBROUTINE new_ncrgcnt
  ! ----------------------------------------------------------------------

  ! ------------------------------------------------------------------
  SUBROUTINE ncrgcnt_halt(substr, status)

    ! MESSy
    USE MESSY_MAIN_CONSTANTS_MEM,  ONLY: STRLEN_VLONG
    USE MESSY_MAIN_GRID_NETCDF,    ONLY: RGMSG, RGMLE
    
    IMPLICIT NONE

    INTRINSIC :: TRIM

    ! I/O
    CHARACTER(LEN=*), INTENT(IN)  :: substr
    INTEGER,          INTENT(IN)  :: status
    ! LOCAL
    CHARACTER(LEN=STRLEN_VLONG)   :: errstr
    
    IF (status == 0) RETURN
    
    errstr = ncrgcnt_error_str(status)
    
    CALL RGMSG(substr, RGMLE, TRIM(errstr))
    
  END SUBROUTINE ncrgcnt_halt
  ! -------------------------------------------------------------------
  
  ! -------------------------------------------------------------------
  FUNCTION ncrgcnt_error_str(status)
    
    USE MESSY_MAIN_TOOLS,         ONLY: int2str
    USE MESSY_MAIN_CONSTANTS_MEM, ONLY: STRLEN_VLONG
    
    IMPLICIT NONE
    
    ! I/O
    CHARACTER(LEN=STRLEN_VLONG)           :: ncrgcnt_error_str
    INTEGER,                   INTENT(IN) :: status
    
    ! LOCAL
    CHARACTER(LEN=4) :: echar = '    '
    
    ncrgcnt_error_str = ''
    
    CALL INT2STR(echar, status, cpad='0', cerr='*')
    
    SELECT CASE(echar)
       ! NO ERROR
    CASE('0000')
       ncrgcnt_error_str = 'E'//echar//': NO ERROR'

    CASE('5003')
       ncrgcnt_error_str = 'E'//echar//': SUBMODEL NAME TOO LONG'
    CASE('5004')
       ncrgcnt_error_str = 'E'//echar//': COUNTER EXISTS ALREADY'
    CASE('5005')
       ncrgcnt_error_str = 'E'//echar//': COUNTER DOES NOT EXIST'
    CASE('5006')
       ncrgcnt_error_str = 'E'//echar//': COUNTER NAME TOO LONG'

    CASE('5020')
       ncrgcnt_error_str = 'E'//echar//': COUNTER FILE NOT FOUND'

    CASE DEFAULT
       ncrgcnt_error_str = 'E'//echar//': UNKONW ERROR STATUS'

    END SELECT

  END FUNCTION ncrgcnt_error_str
  ! -------------------------------------------------------------------


! ******************************************************************
SUBROUTINE READ_CONTROL(GCTRL, GNML, GSTAT              &
                        ,var                            &
                        ,gridin                         &
                        ,gridout                        &
                        ,RG_TYPE                        &
                        ,lint                           &
                        ,nmlfile                        &
                        ,tc, tmin, tmax, tstep, tret    &
                        ,iounit                         &
                        ,ldomainpar                     & 
                        ,lvarparallel                   & 
                        ,lpresaxis                      & 
                        ,lwork                          & 
                        )

  USE MESSY_MAIN_CONSTANTS_MEM, ONLY: dp, sp
  USE MESSY_MAIN_GRID_TRAFO,    ONLY: CHECK_GEOHYBGRID &
                                    , CHECK_NCVAR_ON_GEOHYBGRID &
                                    , REDUCE_INPUT_GRID         &
                                    , RG_IXF ! um_ak_20150217
  USE MESSY_MAIN_GRID,          ONLY: t_geohybgrid, COPY_GEOHYBGRID           &
                                    , IMPORT_GEOHYBGRID, INIT_GEOHYBGRID &
                                    , EXPORT_GEOHYBGRID
  USE MESSY_MAIN_GRID_NETCDF,   ONLY: RGMSG, RGMLE, RGMLEC, RGMLI,  RGMLIC    &
                                    , RGMSG, RGMLW, RGMLWC, RGMLVL, RGMLVM    & 
                                    , RGMLVMC, RGMLVLC, RGMSG, ERRMSG         &
                                    , GRD_MAXSTRLEN, my_rank                  &
                                    , t_ncvar, t_ncatt, t_narray, INIT_NARRAY &
                                    , INIT_NCATT, ADD_NCATT, INIT_NCVAR       &
                                    , IMPORT_NCVAR, RENAME_NCVAR, COPY_NCVAR  &
                                    , SCALE_NARRAY, COPY_NCDIM                &
                                    , NULL_DIMID , nf90_inq_libvers           &
                                    , MAIN_GRID_SET_MESSAGEMODE, IMPORT_NCATT &
                                    , NF90_GLOBAL, VTYPE_UNDEF, VTYPE_CHAR    &
                                    , NF90_CHAR, CAT_NARRAY, MSGMODE_VL       &
                                    , MSGMODE_S, MSGMODE_E

  IMPLICIT NONE

  ! I/O
  INTEGER, INTENT(IN)                   :: GCTRL   ! RG_SCAN, RG_PROC, RG_STOP
  INTEGER, INTENT(IN)                   :: GNML    ! NML_NEXT, NML_STAY
  INTEGER, INTENT(INOUT)                :: GSTAT   ! RGSTAT_START, RGSTAT_CONT
                                                   ! RGSTAT_STOP
  TYPE (t_ncvar), DIMENSION(:), POINTER :: var     ! list of variables
  TYPE (t_geohybgrid), INTENT(INOUT)    :: gridin  ! input grid info
  TYPE (t_geohybgrid), INTENT(INOUT)    :: gridout ! output grid info
  INTEGER, DIMENSION(:),   POINTER      :: RG_TYPE ! regridding type
  LOGICAL, INTENT(OUT)                  :: lint    ! input time ?
  CHARACTER(LEN=*) ,    INTENT(IN)      :: nmlfile ! namelist file
  !
  ! OPTIONAL I/O
  INTEGER          , INTENT(IN) , OPTIONAL :: tc          ! current step
  INTEGER          , INTENT(OUT), OPTIONAL :: tmin        ! start time step
  INTEGER          , INTENT(OUT), OPTIONAL :: tmax        ! stop time step
  INTEGER          , INTENT(OUT), OPTIONAL :: tstep       ! time step
  INTEGER          , INTENT(OUT), OPTIONAL :: tret        ! return time step
  INTEGER          , INTENT(IN) , OPTIONAL :: iounit      ! I/O unit for nmlfile
  LOGICAL          , INTENT(IN),  OPTIONAL :: ldomainpar  ! parallel application
  LOGICAL          , INTENT(IN),  OPTIONAL :: lvarparallel! parallel application

  ! use pressure axis (only required for NCREGRID)
  LOGICAL          , INTENT(OUT), OPTIONAL :: lpresaxis 
  LOGICAL          , INTENT(OUT), OPTIONAL :: lwork     

  ! LOCAL (for DEFAULT)
  CHARACTER(LEN=*), PARAMETER :: substr = 'READ_CONTROL'
  INTEGER                     :: iou   ! I/O unit for namelist file
  ! LOCAL
  LOGICAL                     :: lex   ! file exists ?
  LOGICAL                     :: lopn  ! file opened ?
  INTEGER                     :: iout  ! I/O unit
  INTEGER                     :: i,j   ! counter
  INTEGER                     :: nvars ! number of variables
  INTEGER                     :: mvars ! 'effective' number of variables
  INTEGER                     :: status
  TYPE (t_ncvar), DIMENSION(:), POINTER  :: xivar ! list of variables (input)
  TYPE (t_ncvar)                         :: tvar  ! temporal variable

  ! NAMELIST
  INTEGER, SAVE :: fstat ! status of namelist file
  TYPE (t_geohybgrid)                     :: gi      ! input grid
  TYPE (t_geohybgrid),               SAVE :: gi0     ! input grid (raw)
  TYPE (t_geohybgrid)                     :: gg      ! grdfile grid
  TYPE (t_geohybgrid),               SAVE :: gg0     ! grdfile grid
  CHARACTER (LEN=100*GRD_MAXSTRLEN), SAVE :: varstr  ! variable string
  CHARACTER (LEN=GRD_MAXSTRLEN),     SAVE :: outfile ! netCDF output-file
  INTEGER,                           SAVE :: i_t(4)  ! input time control
  INTEGER,                           SAVE :: g_t(3)  ! grid time control
  INTEGER,                           SAVE :: o_t(3)  ! output time control
  LOGICAL,                           SAVE :: lp      ! pressure or sigma
!!$  TYPE (t_ncatt),SAVE                     :: rggatt  ! RG global attribute ! um_ak_20140328
  TYPE (t_narray),                   SAVE :: nmlstr  ! namelist as string
  !
  ! TIME CONTROL
  INTEGER                                 :: ttc
  INTEGER                                 :: ttmin, ttmax, ttstep, ttret
  INTEGER                                 :: tg       ! grid time step
  INTEGER                                 :: to       ! output time step
  LOGICAL                                 :: ok       ! result OK ?
  LOGICAL                                 :: ldompar = .FALSE. 
  LOGICAL                                 :: lvarpar = .FALSE. 
  ! FORCE UNLIMITED ID
  INTEGER                                 :: setuid

  ! VARIABLES
  CHARACTER(LEN=GRD_MAXSTRLEN), POINTER, SAVE :: ivar(:) ! input variable names
  CHARACTER(LEN=GRD_MAXSTRLEN), POINTER   :: ovar(:)   ! output variable names
  REAL                        , POINTER   :: scl(:)    ! scaling
  INTEGER                     , POINTER   :: RGT(:)    ! regridding type
  CHARACTER (LEN=3)           , POINTER   :: RGTstr(:) ! regridding type string
  INTEGER, DIMENSION(:)       , POINTER   :: dims      ! order of grid dimensions
  INTEGER                                 :: axes(3)   ! grid axes of variable

  INTEGER, DIMENSION(2) :: zstart, zcount

  INTEGER, SAVE       :: GSTAT_SAVE

  IF (PRESENT(ldomainpar)) THEN
     ldompar =ldomainpar
  ELSE
     ldompar = .FALSE.
  ENDIF
  IF (PRESENT(lvarparallel)) THEN
     lvarpar = lvarparallel
  ELSE
     lvarpar = .FALSE.
  ENDIF
  ! parallelise only once (domain  => ldompar = .TRUE. 
  !                     or species => lvarpar = .TRUE. )
  IF (ldompar) lvarpar = .FALSE.

  IF(lvarpar .OR. ldompar)  THEN
     me    = my_mpi%rank
     nproc = my_mpi%nproc
     comm  = my_mpi%comm
  ELSE
     me    = 0
     nproc = 1
     comm  = -1
  END IF

  CALL READ_CONTROL_INIT

  IF (i_am_worker) THEN
     CALL READ_CONTROL_WORK
  END IF

  GSTAT_SAVE = GSTAT

  IF (PRESENT(lwork)) lwork = i_am_worker

  ! um_ak_20140314+
!  CALL INIT_GEOHYBGRID(gi0)
!  CALL INIT_GEOHYBGRID(gg0)
!  CALL INIT_NARRAY(nmlstr)
  ! um_ak_20140314-

  RETURN

 CONTAINS

 SUBROUTINE READ_CONTROL_INIT

  ! INIT
  CALL MAIN_GRID_SET_MESSAGEMODE(MSGMODE_S + MSGMODE_E + MSGMODE_VL)

  ! CHECK INTERFACE
  IF ((GNML /= NML_NEXT).AND.(GNML /= NML_STAY)) THEN
     CALL RGMSG(substr, RGMLE, 'UNRECOGNIZED VALUE OF RG_NML !')
  END IF

  ! SET DEFAULTS
  IF (PRESENT(iounit)) THEN
     iou = iounit
  ELSE
     iou = 17                  ! DEFAULT
  END IF
  !
  ! INITIALIZE OUTPUT PARAMETER
  IF (ASSOCIATED(var)) THEN
     DO i=1, SIZE(var)
        CALL INIT_NCVAR(var(i))
     END DO
     DEALLOCATE(var, STAT=status)
     CALL ERRMSG(substr,status,1)
  END IF
  NULLIFY(var)
  CALL INIT_GEOHYBGRID(gridin)

  ! INIT
  NULLIFY(dims)
  NULLIFY(ivar)
  NULLIFY(ovar)
  NULLIFY(scl)
  NULLIFY(RGT)
  NULLIFY(RGTstr)
  NULLIFY(xivar)

  ! CHECK NAMELIST FILE; OPEN FILE
  INQUIRE(file=TRIM(nmlfile), exist=lex, opened=lopn, number=iout)
  IF (.NOT.lex) THEN
     CALL RGMSG(substr, RGMLE, &
          'NAMELIST-FILE '''//TRIM(nmlfile)//''' NOT FOUND !')
  END IF
  IF (lopn) THEN  ! FILE OPEN
     IF (iout /= iou) THEN   ! CHECK UNIT
        CALL RGMSG(substr, RGMLE, &
             'NAMELIST-FILE '''//TRIM(nmlfile)//''' ALREADY OPEN ON UNIT', &
             iout,' !')
     ELSE
        GSTAT = RGSTAT_CONT
     END IF
  ELSE            ! FILE NOT OPEN
     CALL RGMSG(substr, RGMLVL, &
         '-------------------------------------------------------')
     CALL RGMSG(substr, RGMLVL, 'START REGRIDDING PROCEDURE ')
     CALL RGMSG(substr, RGMLVL, &
          ' <Author: Patrick Joeckel, MPICH, June 2002>' )
     CALL RGMSG(substr, RGMLVL, &
          ' ( COMPILED WITH netCDF Fortran90 LIBRARY' )
     CALL RGMSG(substr, RGMLVL, &
          '   VERSION '//TRIM(NF90_inq_libvers())//' )' )
     CALL RGMSG(substr, RGMLVL, &
          '-------------------------------------------------------')
     CALL RGMSG(substr, RGMLVL, &
          'OPENING NAMELIST FILE '''//TRIM(nmlfile)//''' ON UNIT', &
          iou,' !')
     OPEN(iou,file=TRIM(nmlfile))
     GSTAT = RGSTAT_START
  END IF

  ! READ NEXT NAMELIST
  IF ((GSTAT == RGSTAT_START).OR.(GNML == NML_NEXT)) THEN
     
     ! um_ak_20150720+
     CALL INIT_GEOHYBGRID(gi0)
     CALL INIT_GEOHYBGRID(gg0)
     CALL INIT_NARRAY(nmlstr)
     ! um_ak_20150720-

     CALL RGMSG(substr, RGMLI, &
          'READING NAMELIST FROM FILE '''//TRIM(nmlfile)//''' ...')
     CALL READ_NAMELIST(iou, fstat               &
          ,gi0, gg0, outfile       &
          ,varstr                  &
          ,i_t, g_t, o_t, lp, lint, nmlstr)
     ! CHECK END OF FILE
     IF (fstat == 0) THEN    ! GO ON
        CALL RGMSG(substr, RGMLIC, '... OK !')

        ! GET VARIABLE INFORMATION
        CALL PARSE_VARSTR(varstr, ivar, ovar, scl     &
             ,RGT, RGTstr,TRIM(gi0%file))

        IF (lvarpar .OR. ldompar) THEN
        ! NOTE: THE FOLLOWING SUBROUTINE CALLS ARE REQUIRED TO SET
        !       i_am_worker CORRECTLY
           IF (ldompar) THEN
              i_am_worker = .TRUE.
              ! in parallel domain decomposition mode, each PE must work on
              ! all variables; nproc_work is the number of PEs among which the
              ! number of variables is distributet; thus nproc_work = 1 here.
              nproc_work  = 1
           ELSE
              ! nproc_work and i_am_worker ist set in this subroutine
              CALL DISTRIBUTE_VARS_ON_PES(ivar, ovar, scl, RGT, RGTstr)
              DO i=1, SIZE(ivar)
                 IF (i_am_worker) &
                      CALL RGMSG(substr, RGMLIC &
                      , '... ... '//TRIM(ivar(i))//' on PE=', me, '')
              END DO
           ENDIF
        ELSE
           ! ... in this case, the routine has been called only on p_io ...
           i_am_worker = .TRUE.
           nproc_work  = 1
        ENDIF
        !
        GSTAT = RGSTAT_CONT
        IF (PRESENT(lpresaxis)) lpresaxis = lp ! um_ak_20131018
        !
     ELSE                    ! ERROR OR END OF FILE
        i_am_worker = .FALSE. ! mz_kk_20080218
        IF (fstat > 0) THEN  ! ERROR
           CLOSE(iou)
           CALL RGMSG(substr, RGMLE, &
                'READ ERROR READING NAMELIST FROM FILE '''//&
                &TRIM(nmlfile)//''' !' , .false.)
           CALL RGMSG(substr, RGMLEC, &
                'STATUS: ',fstat,' ')
        ELSE                 ! END OF FILE
           CALL RGMSG(substr, RGMLVL, &
                'END OF FILE '''//TRIM(nmlfile)//''' REACHED !' )
        END IF               ! ERROR OR END OF FILE
        GSTAT = RGSTAT_STOP
     END IF
  ELSE IF (GNML == NML_STAY) THEN
     ! GET VARIABLE INFORMATION
     CALL PARSE_VARSTR(varstr, ivar, ovar, scl     &
          ,RGT, RGTstr,TRIM(gi0%file))
  END IF

  IF (GCTRL == RG_STOP) GSTAT = RGSTAT_STOP

  ! --------------------------------------------------
  IF (GSTAT == RGSTAT_STOP) THEN

     CLOSE(iou)
     ! CLEAN
     CALL RGMSG(substr, RGMLVM, 'CLEANING MEMORY ...')
     CALL INIT_GEOHYBGRID(gridin)
     !
     IF (ASSOCIATED(var)) THEN
        DO i=1, SIZE(var)
           CALL INIT_NCVAR(var(i))
        END DO
        DEALLOCATE(var, STAT=status)
        CALL ERRMSG(substr,status,2)
     END IF
     IF (ASSOCIATED(RG_TYPE)) THEN
        DEALLOCATE(RG_TYPE)
        NULLIFY(RG_TYPE) ! op_pj_20140820
     ENDIF
     NULLIFY(var)
     !
     CALL INIT_GEOHYBGRID(gi0)
     CALL INIT_GEOHYBGRID(gg0)
     CALL INIT_NARRAY(nmlstr)
!!$     CALL INIT_NCATT(rggatt) ! um_ak_20140328
     !
     CALL RGMSG(substr, RGMLVM, '... DONE (CLEANING MEMORY) !')
     !
     ! MESSAGE
     CALL RGMSG(substr, RGMLVL, &
         '-------------------------------------------------------')
     CALL RGMSG(substr, RGMLVL, &
          'END REGRIDDING PROCEDURE')
     CALL RGMSG(substr, RGMLVL, &
         '-------------------------------------------------------')
     RETURN
  END IF
  ! --------------------------------------------------
  RETURN

END SUBROUTINE READ_CONTROL_INIT

SUBROUTINE READ_CONTROL_WORK

  IF (GSTAT == RGSTAT_STOP) RETURN 

  ! --- TIME CONTROL ---------------------------------
  ! UPDATE TIME CONTROL FROM NAMELIST
  ttmin = i_t(1)
  ttstep = i_t(2)
  ttmax = i_t(3)
  ttret = i_t(4)
  !
  ! OVERWRITE TIME CONTROL FROM SUBROUTINE CALL
  IF (PRESENT(tmin)) THEN
     tmin = ttmin
  END IF
  IF (PRESENT(tmax)) THEN
     tmax = ttmax
  END IF
  IF (PRESENT(tstep)) THEN
     tstep = ttstep
  END IF
  IF (PRESENT(tret)) THEN
     tret = ttret
  END IF
  !
  ! CHECK TIME CONTROL SETTINGS
  ! CHECK INFILE TIME STEPPING / INTERFACE MODE TIME STEPPING
  IF (ttmin <= 0) THEN
     CALL RGMSG(substr, RGMLE,  'i_t(1) IN NAMELIST OR tmin AT ', .false.)
     CALL RGMSG(substr, RGMLEC, 'SUBROUTINE CALL MUST BE >0 !')
  END IF
  IF (ttstep <= 0) THEN
     CALL RGMSG(substr, RGMLE,  'i_t(2) IN NAMELIST OR tstep AT ', .false.)
     CALL RGMSG(substr, RGMLEC, 'SUBROUTINE CALL MUST BE >0 !')
  END IF
  IF (ttmax < 0) THEN
     CALL RGMSG(substr, RGMLE,  'i_t(3) IN NAMELIST OR tmax AT ', .false.)
     CALL RGMSG(substr, RGMLEC, 'SUBROUTINE CALL MUST BE >=0 !')
  END IF
  IF (ttret < 0) THEN
     CALL RGMSG(substr, RGMLE,  'i_t(4) IN NAMELIST OR tret AT ', .false.)
     CALL RGMSG(substr, RGMLEC, 'SUBROUTINE CALL MUST BE >=0 !')
  END IF
  !
  IF (ttmax == 0) THEN  ! SPECIAL CASE
     CALL RGMSG(substr, RGMLW,  'tmax (i_t(3)) RESET TO tmin (i_t(1)) !')
     ttmax = ttmin
  END IF
  IF (ttret == 0) THEN  ! SPECIAL CASE
     CALL RGMSG(substr, RGMLW,  'tret (i_t(4)) RESET TO tmin (i_t(1)) !')
     ttret = ttmin
  END IF
  !
  IF (ttmax < ttmin) THEN
     CALL RGMSG(substr, RGMLE,  'i_t(3) IN NAMELIST OR tmax AT ', .false.)
     CALL RGMSG(substr, RGMLEC, 'SUBROUTINE CALL MUST BE >= i_t(1) OR tmin!')
  END IF
  IF ( .NOT.( (ttret >= ttmin).AND.(ttret <= ttmax) ) ) THEN
     CALL RGMSG(substr, RGMLE, &
          'NCREGRID IN INTERFACE MODE ONLY RETURNS DATA, IF', .false.)
     CALL RGMSG(substr, RGMLEC, &
          '   i_t(1) <= i_t(4) <= i_t(3)  IN THE NAMELIST, OR', .false.)
     CALL RGMSG(substr, RGMLEC, &
          '   tmin   <= tret   <= tmax    AT SUBROUTINE CALL!', .false.)
     CALL RGMSG(substr, RGMLEC, &
          'PLEASE CORRECT NAMELIST (MOST PROBABLY i_t(4)), OR', .false.)
     CALL RGMSG(substr, RGMLEC, &
          'PARAMETERS IN SUBROUTINE CALL (tret) ACCORDINGLY!')
  END IF
  ! CHECK GRDFILE TIME STEPPING
  ! NOTE: IN PRINCIPLE IT IS POSSIBLE TO SPECIFY A GRDFILE (IN THE NAMELIST)
  !       IN INTERFACE MODE; THIS OVERWRITES THE GRID SPECIFIED BY THE
  !       INTERFACE.
  IF (g_t(1) <= 0) THEN
     CALL RGMSG(substr, RGMLE, 'g_t(1) IN NAMELIST MUST BE >0 !')
  END IF
  IF (g_t(2) <= 0) THEN
     CALL RGMSG(substr, RGMLE, 'g_t(2) IN NAMELIST MUST BE >0 !')
  END IF
  IF (g_t(3) <= g_t(1)) THEN
     CALL RGMSG(substr, RGMLW, &
          'GRDFILE TIME STEP ',g_t(1),' WILL BE USED')
     CALL RGMSG(substr, RGMLWC, 'FOR ALL INFILE TIME STEPS !')
  END IF
  ! CHECK OUTFILE TIME STEPPING
  IF (o_t(1) <= 0) THEN
     CALL RGMSG(substr, RGMLE, 'o_t(1) IN NAMELIST MUST BE >0 !')
  END IF
  IF (o_t(1) <= 0) THEN
     CALL RGMSG(substr, RGMLE, 'o_t(2) IN NAMELIST MUST BE >0 !')
  END IF
  ! o_t(3) IS NOT USED
  ! --- END TIME CONTROL ---------------------------------

  IF (GNML == NML_NEXT .AND. (ldompar .OR. lvarpar)) THEN
     CALL EXPAND_FILENAME(outfile, me)
  END IF

  CALL RGMSG(substr, RGMLIC, '... OUTPUT FILE '//TRIM(outfile))
  CALL RGMSG(substr, RGMLIC, '... OK !')

  nvars = SIZE(ivar)
  ALLOCATE(xivar(nvars), STAT=status)
  CALL ERRMSG(substr,status,3)

  ! START TIME LOOP
  tg = g_t(1) - g_t(2)              ! INITIALIZE time step in grid file
  to = o_t(1) - o_t(2)              ! INITIALIZE time step in output file

  IF (PRESENT(tc)) THEN
     ttc = tc
  ELSE
     ttc = ttmin
  ENDIF

  CALL RGMSG(substr, RGMLVL, &
       '-------------------------------------------------------')

  ! copy raw grid from namelist input
  CALL COPY_GEOHYBGRID(gi, gi0)
  CALL COPY_GEOHYBGRID(gg, gg0)

  ! UPDATE OUTPUT TIME STEP
  tg = tg + g_t(2)              ! NEXT TIME STEP IN GRID FILE
  to = to + o_t(2)              ! NEXT OUTPUT TIME STEP
  IF ((tg > g_t(3)).AND.(g_t(3) > 0)) THEN  ! RESET
     tg = g_t(1)
  END IF
  gi%t = ttc
  gg%t = tg
  CALL RGMSG(substr, RGMLVM, 'INFILE  TIME STEP: ',gi%t,' ')
  CALL RGMSG(substr, RGMLVM, 'GRDFILE TIME STEP: ',gg%t,' ')
  CALL RGMSG(substr, RGMLVM, 'OUTFILE TIME STEP: ',to,' ')
  IF (lint) THEN
     CALL RGMSG(substr, RGMLVM, ' USING INFILE TIME FOR OUTPUT !')
  ELSE
     CALL RGMSG(substr, RGMLVM, ' USING GRDFILE TIME FOR OUTPUT !')
  END IF
  
  ! GET INPUT GRID
  CALL RGMSG(substr, RGMLVL, &
       'IMPORTING INPUT GRID ... ')
  CALL RGMSG(substr, RGMLVL, &
       '... IMPORTING FROM '''//TRIM(gi%file)//'''')
  CALL IMPORT_GEOHYBGRID(gi)
  !
  CALL RGMSG(substr, RGMLIC, ' ... CHECKING ...')
  CALL CHECK_GEOHYBGRID(gi)
  CALL RGMSG(substr, RGMLIC, ' ... O.K.')
  
  ! GET DESTINATION GRID
  CALL RGMSG(substr, RGMLVL, 'IMPORTING OUTPUT GRID ...')
  IF (gridout%ID == -99) THEN
     CALL RGMSG(substr, RGMLVL, &
          '... IMPORTING FROM '''//TRIM(gg%file)//'''')
     CALL IMPORT_GEOHYBGRID(gg)
     CALL COPY_GEOHYBGRID(gridout, gg)
  END IF

  IF (lint) THEN  ! ADJUST OUTPUT TIME STEPPING FOR INTERFACE MODE
     to = ttc
     gridout%t = gi%t
  ELSE
     to = tg
  END IF

  ! REDUCE INPUT GRID ACCORDING TO OUTPUT GRID     
  CALL RGMSG(substr, RGMLVL, 'REDUCING INPUT GRID ...')
  ! um_ak_20150217+ op_mm_20150217+
  IF (ANY(RGT(:)== RG_IXF)) THEN
     CALL RGMSG(substr, RGMLVL, 'IXF! => SCIPPING GRID REDUCTION ...')
  ELSE
  ! um_ak_20150217- op_mm_20150217-
     CALL REDUCE_INPUT_GRID(gi, gridout)
  END IF   ! um_ak_20150217 op_mm_20150217
  
  ! GET INPUT VAR
  CALL RGMSG(substr, RGMLVL, &
       'IMPORTING ',nvars,' VARIABLE(S) FROM FILE '''//TRIM(gi%file)//'''')
  CALL RGMSG(substr, RGMLVL, &
       '('''//TRIM(gi%timem%name)//''' STEP: ',ttc,')')
  mvars = 0
  
  DO i=1, nvars  ! LOOP OVER VARIABLE NAMES
     CALL RGMSG(substr, RGMLVL,' ... '''//TRIM(ivar(i))//''' ...')
     IF (ASSOCIATED(gi%timem%dim)) THEN
        setuid = gi%timem%dim(1)%id
     ELSE
        setuid = NULL_DIMID
     END IF
     IF (gi%start(1) > 0 .AND. gi%start(2) > 0 .AND. &
          gi%count(1) > 0 .AND. gi%count(2) > 0)  THEN
        zstart = gi%start
        zcount = gi%count
        
        CALL IMPORT_NCVAR(tvar, ttc, varname=TRIM(ivar(i)) &
             ,file=TRIM(gi%file)             &
             ,setuid=setuid                  &
             ,pstart=zstart, pcount=zcount   ) 
        
     ELSE
        CALL IMPORT_NCVAR(tvar, ttc, varname=TRIM(ivar(i)) &
             ,file=TRIM(gi%file)             &
             ,setuid=setuid                  ) 
     ENDIF
 
     ! CHECK
     CALL RGMSG(substr, RGMLIC, '  ... CHECKING FOR GRID CONFORMITY ...')
     CALL CHECK_NCVAR_ON_GEOHYBGRID(tvar, gi, dims, axes, ok)
     
     ! NOT NEEDED HERE
     axes(:) = 0
     DEALLOCATE(dims, STAT=status)
     CALL ERRMSG(substr,status,4)
     NULLIFY(dims)
     !
     IF (ok) THEN
        ! op_pj_20141014+
        ! FILENAME
        CALL ADD_NCATT(tvar, 'mmig_input_file', vs=TRIM(gi%file))
        ! RENAME
        IF (TRIM(ivar(i)) /= TRIM(ovar(i))) THEN
           CALL RGMSG(substr, RGMLVLC, &
                '  ... RENAMING TO '''//TRIM(ovar(i))//'''... ')
           CALL RENAME_NCVAR(tvar, TRIM(ovar(i)))
!!$           CALL ADD_NCATT(tvar, 'RG_ORIGINAL_NAME', vs=TRIM(ivar(i)))
        END IF
        CALL ADD_NCATT(tvar, 'mmig_variable_name', vs=TRIM(ivar(i)))
        ! op_pj_20141014-
        ! SCALE
        IF (ABS(scl(i) - 1.0) >= TINY(1.0)) THEN
           CALL  RGMSG(substr, RGMLVLC,'  ... SCALING BY ',scl(i),' ...')
           CALL SCALE_NARRAY(tvar%dat, scl(i))
!!&           CALL ADD_NCATT(tvar, 'RG_SCALED_BY',      &
           CALL ADD_NCATT(tvar, 'mmig_scaled_by',      &
                vr=(/ REAL(scl(i), SP) /))
        END IF
        ! ADJUST TIME AXIS
        ! if input_time (= lint) == .false.
        ! variable must get correct time-dimension-ID
        IF (.NOT.lint) THEN
           IF (.NOT.(setuid == NULL_DIMID)) THEN
              DO j=1, tvar%ndims
                 IF (tvar%dim(j)%fuid) THEN
                    EXIT
                 END IF
              END DO
              CALL RGMSG(substr, RGMLVMC,'  ... ADJUSTING TIME-DIM.: ')
              CALL RGMSG(substr, RGMLVMC,'      '//TRIM(tvar%dim(j)%name)&
                   //' (ID = ',j,') -> ')
              CALL RGMSG(substr, RGMLVMC,'      '&
                   &//TRIM(gridout%timem%dim(1)%name)&
                   //' (ID = ',setuid,') ...')
              CALL COPY_NCDIM(tvar%dim(j), gridout%timem%dim(1))
           END IF
        END IF
        ! SAVE IN LIST
        mvars = mvars + 1
        CALL COPY_NCVAR(xivar(mvars), tvar)
        
        RGT(mvars)    = RGT(i)      ! shift
        RGTstr(mvars) = RGTstr(i)   ! shift
     ELSE
        CALL  RGMSG(substr, RGMLVMC, &
             '  ... NOT CONSISTENT WITH GRID! IGNORING !')
     END IF
     ! CLEAN
     CALL INIT_NCVAR(tvar)
  END DO        ! LOOP OVER VARIABLE NAMES
  
  ! WHAT TO DO WITH IMPORTED VARIABLES
  ! ***************************************************
  SELECT CASE(GCTRL)
     ! ***************************************************
     ! ---------------------------------------------------
  CASE(RG_SCAN)                ! RETURN INPUT VARIABLES
     ! ---------------------------------------------------
     CALL RGMSG(substr, RGMLVL, 'SCAN MODE ...')
     ! RETURN VAR
     CALL RGMSG(substr, RGMLVM, 'TIME STEP ...',ttret,' ')
     CALL RGMSG(substr, RGMLVM, '... RETURNING RAW DATA')
     ALLOCATE(var(mvars), STAT=status)
     CALL ERRMSG(substr,status,5)
     ALLOCATE(RG_TYPE(mvars), STAT=status)
     CALL ERRMSG(substr,status,5)
     
     DO i=1, mvars
        CALL INIT_NCVAR(var(i))
        CALL COPY_NCVAR(var(i),xivar(i))
        RG_TYPE(i) = RGT(i)
     END DO
     ! RETURN INPUT GRID
     CALL RGMSG(substr, RGMLVM, '... RETURNING INPUT GRID')
     CALL COPY_GEOHYBGRID(gridin, gi)
     CALL RGMSG(substr, RGMLVL, '... DONE (SCAN MODE) !')

     CALL RGMSG(substr, RGMLVL, 'SETTING OUTFILE DATA (SCAN MODE) !')
     IF (TRIM(outfile) /= '') THEN
        gridout%file = TRIM(outfile)    ! name of output-file
        gridout%t = to                  ! output time step

        ! GET OLD ATTRIBUTE
        CALL CAT_NARRAY(gridout%att%dat, nmlstr)
        gridout%att%dat%type = VTYPE_CHAR
        gridout%att%xtype   = NF90_CHAR
        gridout%att%len     = gridout%att%dat%dim(1)
        gridout%att%name    = 'RG_HISTORY'
        gridout%att%varid   = NF90_GLOBAL
     END IF

     CALL RGMSG(substr, RGMLVL, ' DONE SETTING OUTFILE DATA (SCAN MODE) !')
     
     ! ---------------------------------------------------
  CASE(RG_PROC)                ! REGRID
     ! NOTHING TO DO in READ_CONTROL
     ! ---------------------------------------------------
     ! ---------------------------------------------------
  CASE (RG_STOP)
     ! DO NOTHING; THIS POINT IS NOT REACHED
     ! ---------------------------------------------------
     ! ---------------------------------------------------
  CASE DEFAULT
     ! ---------------------------------------------------
     CALL RGMSG(substr, RGMLE, 'VALUE OF RG_CTRL NOT RECOGNIZED !')
     ! ***************************************************
  END SELECT
     ! ***************************************************

  ! CLEAN
  IF (ASSOCIATED(xivar)) THEN
     DO i=1, SIZE(xivar)
        CALL INIT_NCVAR(xivar(i))
     END DO
     ! DO NOT DEALLOCATE/NULLIFY 'xivar' HERE, SINCE IT
     ! HAS BEEN ALLOCATED OUTSIDE THE TIME LOOP !
  END IF
  !
  CALL INIT_GEOHYBGRID(gg)
  CALL INIT_GEOHYBGRID(gi)
  
  ! CLEAN
  IF (ASSOCIATED(ivar)) DEALLOCATE(ivar)
  IF (ASSOCIATED(ovar)) DEALLOCATE(ovar)
  IF (ASSOCIATED(scl))  DEALLOCATE(scl)
  IF (ASSOCIATED(RGT))  DEALLOCATE(RGT)
  IF (ASSOCIATED(RGTstr)) DEALLOCATE(RGTstr)
  NULLIFY(ivar, ovar, scl, RGT, RGTstr)

  IF (ASSOCIATED(xivar)) THEN
     ! CALL INIT_NCVAR(xivar(i)) HAS ALREADY BEEN DONE INSIDE
     ! (AT THE END OF) THE TIME LOOP
     DEALLOCATE(xivar, STAT=status)
     CALL ERRMSG(substr,status,20)
     NULLIFY(xivar)
  END IF

  END SUBROUTINE READ_CONTROL_WORK 

END SUBROUTINE READ_CONTROL


! ******************************************************************
! ------------------------------------------------------------------
SUBROUTINE PARSE_VARSTR(var, ivar, ovar, scl, RGT, RGTstr, file)

  USE MESSY_MAIN_GRID_NETCDF,  ONLY: GRD_MAXSTRLEN, t_ncvar, ERRMSG            &
                                   , RGMSG, RGMLI, RGMLE, RGMLVM, RGMLW, RGMLWC&
                                   , SCAN_NCVAR, INIT_NCVAR, IMPORT_NCVAR      &
                                   , STRING, INIT_NARRAY
  USE MESSY_MAIN_GRID_TRAFO,   ONLY: RG_INT, RG_EXT, RG_IDX, RG_IXF

  IMPLICIT NONE

  ! I/O
  CHARACTER(LEN=*), INTENT(IN)           :: var       ! variable string
  CHARACTER(LEN=GRD_MAXSTRLEN), POINTER  :: ivar(:)   ! input variable names
  CHARACTER(LEN=GRD_MAXSTRLEN), POINTER  :: ovar(:)   ! output variabel names
  REAL                        , POINTER  :: scl(:)    ! scaling factor
  INTEGER                     , POINTER  :: RGT(:)    ! regridding type
  CHARACTER(LEN=3)            , POINTER  :: RGTstr(:) ! regridding type string
  CHARACTER(LEN=*)            , OPTIONAL, INTENT(IN) :: file      ! filename

  ! LOCAL
  CHARACTER(LEN=*), PARAMETER           :: substr = 'PARSE_VARSTR'
  TYPE (t_ncvar), DIMENSION(:), POINTER :: kvar   ! variables from file
  INTEGER                               :: nvar   ! number of variables
  INTEGER                               :: status
  INTEGER                               :: i,j
  INTEGER                               :: idx1, idx2, idx3
  INTEGER                               :: idxl, idxr
  CHARACTER(GRD_MAXSTRLEN), ALLOCATABLE :: lvar(:)   ! LIST OF VARIABLE INFOs
  INTEGER                               :: len
  CHARACTER(GRD_MAXSTRLEN)              :: infile    ! filename

  ! INIT
  IF (PRESENT(file)) THEN
     infile = TRIM(file)
  ELSE
     infile = ''
  END IF
  NULLIFY(kvar)

  IF (ASSOCIATED(ivar))   DEALLOCATE(ivar);   NULLIFY(ivar)
  IF (ASSOCIATED(ovar))   DEALLOCATE(ovar);   NULLIFY(ovar)
  IF (ASSOCIATED(scl))    DEALLOCATE(scl);    NULLIFY(scl)
  IF (ASSOCIATED(RGT))    DEALLOCATE(RGT);    NULLIFY(RGT)
  IF (ASSOCIATED(RGTstr)) DEALLOCATE(RGTstr); NULLIFY(RGTstr)

  !  count number of variables
  IF ((TRIM(var) == '').OR.(TRIM(var) == ';')) THEN   ! EMPTY VAR LIST
     ! NO VARIABLES GIVEN -> SCAN FOR ALL
     nvar = 0
     CALL RGMSG(substr, RGMLI, 'NO VARIABLES LISTED !')
     IF (TRIM(infile) == '') THEN
        CALL RGMSG(substr, RGMLE,'NO INPUT FILE GIVEN !')
     ELSE
        CALL RGMSG(substr, RGMLVM, &
             'SCANNING INPUT FILE '''//TRIM(infile)//''' ...')
        CALL SCAN_NCVAR(kvar, file=TRIM(infile))
        nvar = SIZE(kvar)
        ALLOCATE(ivar(nvar), STAT=status)
        CALL ERRMSG(substr,status,1)
        ALLOCATE(ovar(nvar), STAT=status)
        CALL ERRMSG(substr,status,2)
        ALLOCATE(scl(nvar), STAT=status)
        CALL ERRMSG(substr,status,3)
        ALLOCATE(RGT(nvar), STAT=status)
        CALL ERRMSG(substr,status,4)
        ALLOCATE(RGTstr(nvar), STAT=status)
        CALL ERRMSG(substr,status,5)
        !
        ! INIT
        scl(:) = 1.0
        !
        DO i=1, nvar
           ivar(i) = TRIM(kvar(i)%name)
           ovar(i) = TRIM(kvar(i)%name)
           ! TRY TO GET RGTstr, RGT from attributes
           RGTstr(i) = '...'
           DO j=1, kvar(i)%natts
              IF (TRIM(kvar(i)%att(j)%name) == 'RG_TYPE') THEN
                 RGTstr(i) = string(kvar(i)%att(j)%dat%vc)
              END IF
           END DO
        END DO
        ! CLEAN UP
        DO i=1, nvar
           CALL INIT_NCVAR(kvar(i))
        END DO
        DEALLOCATE(kvar, STAT=status)
        CALL ERRMSG(substr,status,6)
        NULLIFY(kvar)
     END IF

  ELSE     ! VAR LIST NOT EMPTY

     ! COUNT VARIABLES
     nvar = 1
     len = LEN_TRIM(var)
     DO i=1, len-1        ! last ';' optional !
        IF (var(i:i) == ';') nvar = nvar + 1
     END DO
     ! ALLOCATE SPACE
     ALLOCATE(ivar(nvar), STAT=status)
     CALL ERRMSG(substr,status,7)
     ALLOCATE(ovar(nvar), STAT=status)
     CALL ERRMSG(substr,status,8)
     ALLOCATE(scl(nvar), STAT=status)
     CALL ERRMSG(substr,status,9)
     ALLOCATE(RGT(nvar), STAT=status)
     CALL ERRMSG(substr,status,10)
     ALLOCATE(RGTstr(nvar), STAT=status)
     CALL ERRMSG(substr,status,11)
     ALLOCATE(lvar(nvar), STAT=status)
     CALL ERRMSG(substr,status,12)
     !
     ! INIT
     scl(:) = 1.0
     !
     ! PARSING
     idx1 = 1
     idx2 = INDEX(TRIM(var), ';')
     DO i=1, nvar-1
        lvar(i) = var(idx1:idx2-1)
        idx1 = idx2+1
        idx2 = idx2+INDEX(var(idx1:),';')
     END DO
     IF (idx2 == (idx1-1)) idx2 = len+1    ! last ';' optional !
     lvar(nvar) = var(idx1:idx2-1)
     ! fill ovar, ivar, scl, RGT, RGTstr
     DO i=1, nvar  ! LOOP OVER ALL VARIABLES
        ! GET indices
        len  = LEN_TRIM(lvar(i))
        idx1 = INDEX(lvar(i),'=')
        IF (idx1 == 0) idx1 = len+1
        idx2 = INDEX(lvar(i),',')
        IF (idx2 == 0) idx2 = len+1
        idx3 = INDEX(lvar(i),':')
        IF (idx3 == 0) idx3 = len+1
        ! CHECK for destination variable name
        idxl = 1
        idxr = MIN(idx1, idx2, idx3)-1
        ovar(i) = lvar(i)(idxl:idxr)
        ovar(i) = ADJUSTL(TRIM(ovar(i))) ! op_pj_20130701
        ! CHECK for source variable name
        idxl = idx1+1
        idxr = MIN(idx2, idx3)-1
        IF (idxl > len) idxl = 1
        IF (idxr > len) idxr = len
        ivar(i) = lvar(i)(idxl:idxr)
        ivar(i) = ADJUSTL(TRIM(ivar(i))) ! op_pj_20130701
        ! CHECK for scaling
        idxl = idx2+1
        idxr = idx3-1
        IF (idxr < idxl) idxr = len
        IF (idxl > len) THEN
           scl(i) = 1.0
        ELSE
           READ(lvar(i)(idxl:idxr),*) scl(i)
        END IF
        ! CHECK for RE-GRIDDING TYPE
        idxl = idx3+1
        idxr = idx2-1
        IF (idxr < idxl) idxr = len
        IF (idxl > len) THEN        ! RG_TYPE NOT specified
           RGTstr(i) = '...'
           IF (TRIM(infile) /= '') THEN  ! filename OK
              ALLOCATE(kvar(1), STAT=status)
              CALL ERRMSG(substr,status,13)
              CALL IMPORT_NCVAR(kvar(1), 1, &
                   varname=TRIM(ivar(i)), file=TRIM(infile))
              CALL INIT_NARRAY(kvar(1)%dat)
              DO j=1, kvar(1)%natts
                 IF (TRIM(kvar(1)%att(j)%name) == 'RG_TYPE') THEN
                    RGTstr(i) = string(kvar(1)%att(j)%dat%vc)
                 END IF
              END DO
              ! CLEAN UP
              CALL INIT_NCVAR(kvar(1))
              DEALLOCATE(kvar, STAT=status)
              CALL ERRMSG(substr,status,14)
              NULLIFY(kvar)
           ELSE                          ! filename not OK
              CALL RGMSG(substr, RGMLW, &
                   'REGRIDDING TYPE OF VARIABLE '''//TRIM(ivar(i))//'''')
              CALL RGMSG(substr, RGMLWC, &
                   'CANNOT BE DETERMINED FROM FILE, ')
              CALL RGMSG(substr, RGMLWC, &
                   'SINCE FILENAME IS EMPTY/MISSING !')
           ENDIF                         ! filename
        ELSE                      ! RG_TYPE specified
           RGTstr(i) = lvar(i)(idxl:idxr)
           RGTstr(i) = ADJUSTL(TRIM(RGTstr(i))) ! op_pj_20130701
        END IF                    ! RG_TYPE (NOT) specified
     END DO  ! LOOP OVER ALL VARIABLES

     ! CLEAN
     IF (ALLOCATED(lvar)) THEN
        DEALLOCATE(lvar, STAT=status)
        CALL ERRMSG(substr,status,15)
     END IF
  END IF   ! VAR LIST EMPTY OR NOT

  DO i=1, nvar
     ! CHECK for RE-GRIDDING TYPE
     SELECT CASE (RGTstr(i))
     CASE ('INT','int')
        RGT(i) = RG_INT
     CASE ('EXT','ext')
        RGT(i) = RG_EXT
     CASE ('IDX','idx')
        RGT(i) = RG_IDX
     CASE ('IXF','ixf')
        RGT(i) = RG_IXF
     CASE DEFAULT
        CALL RGMSG(substr, RGMLW, &
             'UNKNOWN RE-GRIDDING TYPE OF VARAIBLE '''//TRIM(ivar(i))//'''')
        CALL RGMSG(substr, RGMLWC, &
             'USING DEFAULT: INT !')
        RGTstr(i) = 'INT'
        RGT(i) = RG_INT
     END SELECT
  END DO

END SUBROUTINE PARSE_VARSTR
! ------------------------------------------------------------------

! ------------------------------------------------------------------
SUBROUTINE READ_NAMELIST(iou, status                    &
                         ,gi, gg, outfile               &
                         ,var, i_t, g_t, o_t, lp, lint  &
                         ,nmlstr)

  USE MESSY_MAIN_CONSTANTS_MEM,  ONLY: dp
  USE MESSY_MAIN_GRID_TRAFO,     ONLY: RGEMPTY, GRIDTRAFOVERS
  USE MESSY_MAIN_GRID_NETCDF,    ONLY: GRD_MAXSTRLEN, t_narray, RGMSG &
                                     , RGMLW, RGMLE, RGMLI, RGMLIC  &
                                     , INIT_NARRAY 
  USE MESSY_MAIN_GRID,           ONLY: t_geohybgrid, INIT_GEOHYBGRID

  IMPLICIT NONE

  ! I/O
  INTEGER                      , INTENT(IN)  :: iou    ! logical I/O unit
  INTEGER                      , INTENT(OUT) :: status ! file status
  TYPE (t_geohybgrid)          , INTENT(OUT) :: gi       ! input grid
  TYPE (t_geohybgrid)          , INTENT(OUT) :: gg       ! output grid
  CHARACTER (LEN=GRD_MAXSTRLEN), INTENT(OUT) :: outfile  ! output filename
  CHARACTER (LEN=100*GRD_MAXSTRLEN), INTENT(OUT) :: var  ! variable string
  INTEGER                      , INTENT(OUT) :: i_t(4)   ! input time control
  INTEGER                      , INTENT(OUT) :: g_t(3)   ! grid time control
  INTEGER                      , INTENT(OUT) :: o_t(3)   ! output time control
  LOGICAL                      , INTENT(OUT) :: lp       ! pressure or sigma
  LOGICAL                      , INTENT(OUT) :: lint     ! input time ?
  TYPE (t_narray)              , INTENT(OUT) :: nmlstr   ! namelist as string

  ! LOCAL
  CHARACTER(LEN=*), PARAMETER :: substr = 'READ_NAMELIST'
  CHARACTER(GRD_MAXSTRLEN) :: infile            ! input netCDF filename
! CHARACTER(GRD_MAXSTRLEN) :: outfile           ! output netCDF filename
  CHARACTER(GRD_MAXSTRLEN) :: grdfile           ! grid file (netCDF)
  CHARACTER(GRD_MAXSTRLEN) :: i_latm, i_lonm    ! input latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: i_lati, i_loni    ! input latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: g_latm, g_lonm    ! output latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: g_lati, g_loni    ! output latitude/longitude
! um_ak_20130730+
  ! rotated coordinates
  CHARACTER(GRD_MAXSTRLEN) :: i_rlatm, i_rlonm    ! input latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: i_rlati, i_rloni    ! input latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: g_rlatm, g_rlonm    ! output latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: g_rlati, g_rloni    ! output latitude/longitude
  ! curvilinear (geographical) coordinates
  CHARACTER(GRD_MAXSTRLEN) :: i_clatm, i_clonm    ! input latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: i_clati, i_cloni    ! input latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: g_clatm, g_clonm    ! output latitude/longitude
  CHARACTER(GRD_MAXSTRLEN) :: g_clati, g_cloni    ! output latitude/longitude
! um_ak_20130730-
  CHARACTER(GRD_MAXSTRLEN) :: i_hyai, i_hybi    ! input hybrid coeff.
  CHARACTER(GRD_MAXSTRLEN) :: i_hyam, i_hybm    ! input hybrid coeff. (mid)
  CHARACTER(GRD_MAXSTRLEN) :: g_hyai, g_hybi    ! output hybrid coeff.
  CHARACTER(GRD_MAXSTRLEN) :: g_hyam, g_hybm    ! output hybrid coeff. (mid)
  CHARACTER(GRD_MAXSTRLEN) :: i_timei, i_timem  ! input time
  CHARACTER(GRD_MAXSTRLEN) :: g_timei, g_timem  ! output time
  CHARACTER(GRD_MAXSTRLEN) :: i_ps, g_ps        ! input/output surface press.
  CHARACTER(GRD_MAXSTRLEN) :: i_p0, g_p0        ! input/output ref. press.
  CHARACTER(GRD_MAXSTRLEN) :: i_pollon, g_pollon ! input/output ref. press.
  CHARACTER(GRD_MAXSTRLEN) :: i_pollat, g_pollat ! input/output ref. press.
  CHARACTER(GRD_MAXSTRLEN) :: i_polgam, g_polgam ! input/output ref. press.
  LOGICAL                  :: pressure          ! pressure or sigma
  LOGICAL                  :: input_time        ! input time information
  !
  REAL(dp), DIMENSION(2)   :: i_latr, i_lonr
  REAL(dp), DIMENSION(2)   :: i_hyar, i_hybr
  LOGICAL                  :: i_lonc         ! um_ak_20130528

  REAL(dp), DIMENSION(2)   :: i_rlatr, i_rlonr 
  LOGICAL                  :: i_rlonc          
  REAL(dp), DIMENSION(4)   :: i_clatr, i_clonr 
  LOGICAL                  :: i_clonc          

  REAL(dp), DIMENSION(2)   :: g_latr, g_lonr
  REAL(dp), DIMENSION(2)   :: g_hyar, g_hybr
  LOGICAL                  :: g_lonc         ! um_ak_20130528

  REAL(dp), DIMENSION(2)   :: g_rlatr, g_rlonr 
  LOGICAL                  :: g_rlonc          
  REAL(dp), DIMENSION(4)   :: g_clatr, g_clonr 
  LOGICAL                  :: g_clonc          
  
  ! HELPERS
  CHARACTER(GRD_MAXSTRLEN) :: tstr
  CHARACTER(LEN=8)         :: date
  CHARACTER(LEN=10)        :: time
  CHARACTER(LEN=5)         :: zone

  NAMELIST /regrid/ infile, outfile, grdfile                           &
       ,i_latm, i_lati, i_lonm, i_loni, i_hyam, i_hyai, i_hybm, i_hybi &
       ,i_timei, i_timem, i_ps, i_p0                                   &
       ,g_latm, g_lati, g_lonm, g_loni, g_hyam, g_hyai, g_hybm, g_hybi &
       ,g_timei, g_timem, g_ps, g_p0                                   &
       ,var, i_t, g_t, o_t, pressure, input_time                       &
       ,i_lonc, i_latr, i_lonr, i_hyar, i_hybr                         &
       ,g_lonc, g_latr, g_lonr, g_hyar, g_hybr                         &
       ,i_rlatm, i_rlati, i_rlonm, i_rloni, g_rlatm, g_rlati, g_rlonm, g_rloni &
       ,i_clatm, i_clati, i_clonm, i_cloni, g_clatm, g_clati, g_clonm, g_cloni &
       ,i_rlonc, g_rlonc,i_clonc, g_clonc &
       , i_pollon, i_pollat, i_polgam, g_pollon, g_pollat, g_polgam

  ! INIT (DEFAULT VALUES)
  infile = ''
  outfile = ''
  grdfile = ''
  i_latm = ''
  i_lati = ''
  i_lonm = ''
  i_loni = ''
  i_hyam = ''
  i_hyai = ''
  i_hybm = ''
  i_hybi = ''
  i_timei = ''
  i_timem = ''
  i_ps = ''
  i_p0 = ''

  i_rlatm = ''
  i_rlati = ''
  i_rlonm = ''
  i_rloni = ''
  i_clatm = ''
  i_clati = ''
  i_clonm = ''
  i_cloni = ''
  i_pollon = ''
  i_pollat = ''
  i_polgam = ''

  g_latm = ''
  g_lati = ''
  g_lonm = ''
  g_loni = ''
  g_hyam = ''
  g_hyai = ''
  g_hybm = ''
  g_hybi = ''
  g_timei = ''
  g_timem = ''
  g_ps = ''
  g_p0 = ''

  g_rlatm = ''
  g_rlati = ''
  g_rlonm = ''
  g_rloni = ''
  g_clatm = ''
  g_clati = ''
  g_clonm = ''
  g_cloni = ''
  g_pollon = ''
  g_pollat = ''
  g_polgam = ''

  var = ''
  i_t(:) = (/ 1,1,0,0 /)
  g_t(:) = (/ 1,1,0 /)
  o_t(:) = (/ 1,1,0 /)
  pressure   = .FALSE.  ! DEFAULT IS: SIGMA LEVELS
  input_time = .TRUE.   ! DEFAULT IS: USE INPUT TIME INFO FOR OUTPUT

  i_lonc    = .TRUE.   
  i_latr(:) = RGEMPTY
  i_lonr(:) = RGEMPTY
  i_hyar(:) = RGEMPTY
  i_hybr(:) = RGEMPTY
  g_lonc    = .TRUE.  
  g_latr(:) = RGEMPTY
  g_lonr(:) = RGEMPTY
  g_hyar(:) = RGEMPTY
  g_hybr(:) = RGEMPTY

  i_rlonc    = .TRUE. 
  g_rlonc    = .TRUE.
  i_rlonc    = .TRUE.
  g_clonc    = .TRUE.

  CALL INIT_GEOHYBGRID(gi)
  CALL INIT_GEOHYBGRID(gg)

  ! READ NAMELIST
  READ(iou, NML=regrid, IOSTAT=status)
  IF (status /= 0) THEN
     RETURN
  END IF

  ! TIME CONTROL CORRECTIONS
  IF (i_t(2) == 0) THEN
     CALL RGMSG(substr, RGMLW, 'INFILE TIME STEP IS ZERO! RESET TO 1!')
     i_t(2) = 1
  END IF
  IF (g_t(2) == 0) THEN
     CALL RGMSG(substr, RGMLW, 'GRDFILE TIME STEP IS ZERO! RESET TO 1!')
     g_t(2) = 1
  END IF
  IF (o_t(2) == 0) THEN
     CALL RGMSG(substr, RGMLW, 'OUTFILE TIME STEP IS ZERO! RESET TO 1!')
     o_t(2) = 1
  END IF

  ! SAVE NAMELIST IN N-ARRAY
  CALL INIT_NARRAY(nmlstr)
  CALL ADDLINE_CNARRAY(nmlstr, 'IMPORT_GRID VERSION '//TRIM(GRIDTRAFOVERS))
  CALL DATE_AND_TIME(date, time, zone)
  CALL ADDLINE_CNARRAY(nmlstr, 'RG_DATE: '//TRIM(date))
  CALL ADDLINE_CNARRAY(nmlstr, 'RG_TIME: '//TRIM(time)//TRIM(zone))
  CALL ADDLINE_CNARRAY(nmlstr, 'NAMELIST:')
  CALL ADDLINE_CNARRAY(nmlstr, '----------------------------------')
  CALL ADDLINE_CNARRAY(nmlstr, 'infile  = '''//TRIM(infile)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'outfile = '''//TRIM(outfile)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'grdfile = '''//TRIM(grdfile)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_latm  = '''//TRIM(i_latm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_lati  = '''//TRIM(i_lati)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_lonm  = '''//TRIM(i_lonm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_loni  = '''//TRIM(i_loni)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_hyam  = '''//TRIM(i_hyam)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_hyai  = '''//TRIM(i_hyai)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_hybm  = '''//TRIM(i_hybm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_hybi  = '''//TRIM(i_hybi)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_timei = '''//TRIM(i_timei)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_timem = '''//TRIM(i_timem)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_ps    = '''//TRIM(i_ps)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_p0    = '''//TRIM(i_p0)//''',')

  CALL ADDLINE_CNARRAY(nmlstr, 'i_rlatm  = '''//TRIM(i_rlatm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_rlati  = '''//TRIM(i_rlati)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_rlonm  = '''//TRIM(i_rlonm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_rloni  = '''//TRIM(i_rloni)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_clatm  = '''//TRIM(i_clatm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_clati  = '''//TRIM(i_clati)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_clonm  = '''//TRIM(i_clonm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_cloni  = '''//TRIM(i_cloni)//''',')

  CALL ADDLINE_CNARRAY(nmlstr, 'g_latm  = '''//TRIM(g_latm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_lati  = '''//TRIM(g_lati)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_lonm  = '''//TRIM(g_lonm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_loni  = '''//TRIM(g_loni)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_hyam  = '''//TRIM(g_hyam)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_hyai  = '''//TRIM(g_hyai)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_hybm  = '''//TRIM(g_hybm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_hybi  = '''//TRIM(g_hybi)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_timei = '''//TRIM(g_timei)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_timem = '''//TRIM(g_timem)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_ps    = '''//TRIM(g_ps)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_p0    = '''//TRIM(g_p0)//''',')

  CALL ADDLINE_CNARRAY(nmlstr, 'g_rlatm  = '''//TRIM(g_rlatm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_rlati  = '''//TRIM(g_rlati)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_rlonm  = '''//TRIM(g_rlonm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_rloni  = '''//TRIM(g_rloni)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_clatm  = '''//TRIM(g_clatm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_clati  = '''//TRIM(g_clati)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_clonm  = '''//TRIM(g_clonm)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_cloni  = '''//TRIM(g_cloni)//''',')

  CALL ADDLINE_CNARRAY(nmlstr, 'var     = '''//TRIM(var)//''',')
  !
  WRITE(tstr,'(4(I6,",",1x))') i_t
  CALL ADDLINE_CNARRAY(nmlstr, 'i_t     = '//TRIM(tstr))
  WRITE(tstr,'(3(I6,",",1x))') g_t
  CALL ADDLINE_CNARRAY(nmlstr, 'g_t     = '//TRIM(tstr))
  WRITE(tstr,'(3(I6,",",1x))') o_t
  CALL ADDLINE_CNARRAY(nmlstr, 'o_t     = '//TRIM(tstr))
  !
  IF (pressure) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'pressure= .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'pressure= .FALSE.')
  END IF
  IF (input_time) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'input_time= .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'input_time= .FALSE.')
  END IF
  !
  IF (i_lonc) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'i_lonc  = .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'i_lonc  = .FALSE.')
  ENDIF
     
  WRITE(tstr,'(e10.4,",",e10.4)') i_lonr(1),i_lonr(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'i_lonr  = '//TRIM(tstr))
  WRITE(tstr,'(e10.4,",",e10.4)') i_latr(1),i_latr(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'i_latr  = '//TRIM(tstr))
  WRITE(tstr,'(e10.4,",",e10.4)') i_hyar(1),i_hyar(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'i_hyar  = '//TRIM(tstr))
  WRITE(tstr,'(e10.4,",",e10.4)') i_hybr(1),i_hybr(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'i_hybr  = '//TRIM(tstr))
  !
  IF (i_rlonc) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'i_rlonc  = .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'i_rlonc  = .FALSE.')
  ENDIF
     
  IF (i_clonc) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'i_clonc  = .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'i_clonc  = .FALSE.')
  ENDIF
     
  CALL ADDLINE_CNARRAY(nmlstr, 'i_pollon = '''//TRIM(i_pollon)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_pollat = '''//TRIM(i_pollat)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'i_polgam = '''//TRIM(i_polgam)//''',')

  IF (g_lonc) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'g_lonc  = .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'g_lonc  = .FALSE.')
  ENDIF
  WRITE(tstr,'(e10.4,",",e10.4)') g_lonr(1),g_lonr(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'g_lonr  = '//TRIM(tstr))
  WRITE(tstr,'(e10.4,",",e10.4)') g_latr(1),g_latr(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'g_latr  = '//TRIM(tstr))
  WRITE(tstr,'(e10.4,",",e10.4)') g_hyar(1),g_hyar(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'g_hyar  = '//TRIM(tstr))
  WRITE(tstr,'(e10.4,",",e10.4)') g_hybr(1),g_hybr(2)
  CALL ADDLINE_CNARRAY(nmlstr, 'g_hybr  = '//TRIM(tstr))
  !
  IF (g_rlonc) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'g_rlonc  = .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'g_rlonc  = .FALSE.')
  ENDIF

  IF (g_clonc) THEN
     CALL ADDLINE_CNARRAY(nmlstr, 'g_clonc  = .TRUE.')
  ELSE
     CALL ADDLINE_CNARRAY(nmlstr, 'g_clonc  = .FALSE.')
  ENDIF

  CALL ADDLINE_CNARRAY(nmlstr, 'g_pollon = '''//TRIM(g_pollon)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_pollat = '''//TRIM(g_pollat)//''',')
  CALL ADDLINE_CNARRAY(nmlstr, 'g_polgam = '''//TRIM(g_polgam)//''',')

  CALL ADDLINE_CNARRAY(nmlstr, '----------------------------------')

  ! CHECK FILENAMES
  IF (TRIM(infile) == '') THEN
     CALL RGMSG(substr, RGMLE, 'INFILE REQUIRED IN NAMELIST !')
  END IF
  IF (TRIM(outfile) == '') THEN
     CALL RGMSG(substr, RGMLW, 'OUTFILE MISSING IN NAMELIST ! NO OUTPUT !')
  END IF
  IF (TRIM(grdfile) == '') THEN
     CALL RGMSG(substr, RGMLI, &
          'GRDFILE MISSING IN NAMELIST ! INTERFACE MODUS: ')
     CALL RGMSG(substr, RGMLIC, &
          'g_lati, g_latm, g_loni, g_lonm,')
     CALL RGMSG(substr, RGMLIC, &
          'g_hyai, g_hyam, g_hybi, g_hybm,')
     CALL RGMSG(substr, RGMLIC, &
          'g_timei, g_timem, g_ps, g_p0, g_t')
     CALL RGMSG(substr, RGMLIC, 'IGNORED FROM NAMELIST !')
  END IF

  ! TRANSFER INFORMATIONS
  gi%file       = TRIM(infile)
  gi%latm%name  = TRIM(i_latm)
  gi%lati%name  = TRIM(i_lati)
  gi%lonm%name  = TRIM(i_lonm)
  gi%loni%name  = TRIM(i_loni)
  gi%hyam%name  = TRIM(i_hyam)
  gi%hyai%name  = TRIM(i_hyai)
  gi%hybm%name  = TRIM(i_hybm)
  gi%hybi%name  = TRIM(i_hybi)
  gi%timei%name = TRIM(i_timei)
  gi%timem%name = TRIM(i_timem)
  gi%ps%name    = TRIM(i_ps)
  gi%p0%name    = TRIM(i_p0)

  gi%rlatm%name  = TRIM(i_rlatm)
  gi%rlati%name  = TRIM(i_rlati)
  gi%rlonm%name  = TRIM(i_rlonm)
  gi%rloni%name  = TRIM(i_rloni)
  gi%clatm%name  = TRIM(i_clatm)
  gi%clati%name  = TRIM(i_clati)
  gi%clonm%name  = TRIM(i_clonm)
  gi%cloni%name  = TRIM(i_cloni)
  gi%pollon%name = TRIM(i_pollon)
  gi%pollat%name = TRIM(i_pollat)
  gi%polgam%name = TRIM(i_polgam)

  gg%file       = TRIM(grdfile)
  gg%latm%name  = TRIM(g_latm)
  gg%lati%name  = TRIM(g_lati)
  gg%lonm%name  = TRIM(g_lonm)
  gg%loni%name  = TRIM(g_loni)
  gg%hyam%name  = TRIM(g_hyam)
  gg%hyai%name  = TRIM(g_hyai)
  gg%hybm%name  = TRIM(g_hybm)
  gg%hybi%name  = TRIM(g_hybi)
  gg%timei%name = TRIM(g_timei)
  gg%timem%name = TRIM(g_timem)
  gg%ps%name    = TRIM(g_ps)
  gg%p0%name    = TRIM(g_p0)

  gg%rlatm%name  = TRIM(g_rlatm)
  gg%rlati%name  = TRIM(g_rlati)
  gg%rlonm%name  = TRIM(g_rlonm)
  gg%rloni%name  = TRIM(g_rloni)
  gg%clatm%name  = TRIM(g_clatm)
  gg%clati%name  = TRIM(g_clati)
  gg%clonm%name  = TRIM(g_clonm)
  gg%cloni%name  = TRIM(g_cloni)
  gg%pollon%name = TRIM(g_pollon)
  gg%pollat%name = TRIM(g_pollat)
  gg%polgam%name = TRIM(g_polgam)

  gi%lonc = i_lonc
  gg%lonc = g_lonc

  gi%rlonc = i_rlonc
  gg%rlonc = g_rlonc
  gi%clonc = i_clonc
  gg%clonc = g_clonc

  lp   = pressure
  lint = input_time

!  var
!  i_t
!  g_t
!  o_t
!  outfile

  gi%ranges(1,:) = i_lonr(:)
  gi%ranges(2,:) = i_latr(:)
  gi%ranges(3,:) = i_hyar(:)
  gi%ranges(4,:) = i_hybr(:)
  gg%ranges(1,:) = g_lonr(:)
  gg%ranges(2,:) = g_latr(:)
  gg%ranges(3,:) = g_hyar(:)
  gg%ranges(4,:) = g_hybr(:)

END SUBROUTINE READ_NAMELIST
! ------------------------------------------------------------------

! ------------------------------------------------------------------
SUBROUTINE ADDLINE_CNARRAY(na, str)

  USE MESSY_MAIN_GRID_NETCDF, ONLY: t_narray, INIT_NARRAY, COPY_NARRAY &
                                  , VTYPE_UNDEF, VTYPE_CHAR, RGMSG   &
                                  , RGMLE

  IMPLICIT NONE

  ! I/O
  TYPE (t_narray),  INTENT(INOUT) :: na
  CHARACTER(LEN=*), INTENT(IN)    :: str

  ! LOCAL
  CHARACTER(LEN=*), PARAMETER :: substr = 'ADDLINE_CNARRAY'
  INTEGER                     :: n, m ! string length
  INTEGER                     :: i    ! counter
  TYPE (t_narray)             :: nas

  ! LOCAL
  IF (na%type == VTYPE_UNDEF) THEN
     ! NEW
     n = LEN_TRIM(str)
     CALL INIT_NARRAY(na, 1, (/ n+1 /), VTYPE_CHAR)   ! +1 for 'NEWLINE'
     DO i=1, n
        na%vc(i) = str(i:i)
     END DO
     na%vc(n+1) = ACHAR(10)
  ELSE
     IF (na%type == VTYPE_CHAR) THEN
        ! SAVE OLD
        CALL INIT_NARRAY(nas)
        CALL COPY_NARRAY(nas, na)
        n = LEN_TRIM(str)
        m = SIZE(nas%vc)
        CALL INIT_NARRAY(na, 1, (/ m+n+1 /), VTYPE_CHAR ) ! +1 for 'NEWLINE'
        DO i=1, m
           na%vc(i) = nas%vc(i)
        END DO
        DO i=1, n
           na%vc(m+i) = str(i:i)
        END DO
        na%vc(n+m+1) = ACHAR(10)
        CALL INIT_NARRAY(nas)
     ELSE
        CALL RGMSG(substr, RGMLE, 'N-ARRAY MUST BE OF TYPE CHAR !')
     END IF
  END IF

END SUBROUTINE ADDLINE_CNARRAY
! ------------------------------------------------------------------
! ******************************************************************

  SUBROUTINE RGTOOL_READ_NCVAR(status, iou, nmlfile, vname, t, var  &
                              , iipol, ogridid, igridid, SCRIP_ID   &
                              , lrg, lrgx, lrgy, lrgz, lok , oarea   &
                              , convgrid, ldompar, lvarpar)

    ! PERFORMS ONE RE-GRIDDING (TIME-)STEP (t) FOR ONE VARIABLE
    ! (vname) AND RETURNS THE RESULT OF TYPE NCVAR (var)
    ! AND OPTIONALLY THE GRID (grid)
    ! LOGICAL I/O UNIT  : iou          (for namelist-file)
    ! NAMELIST-FILE     : nmlfile
    ! DIMENSION-SWITCHES: lrgx, lrgy, lrgz
    ! REGRIDDING-SWITCH : lrg
    ! SUCCESS-FLAG      : lok
    !
    ! Author: Patrick Joeckel, MPICH, October 2002
    !         Astrid Kerkweg,  UniMz, 2013 (restrctured + expanded)

    USE MESSY_MAIN_TOOLS,            ONLY: INT2STR
    ! NCREGRID INTERFACE
    USE MESSY_MAIN_GRID_TRAFO,       ONLY: GTRF_NONE, GTRF_NRGD, GTRF_SCRP
    USE MESSY_MAIN_GRID_TRAFO_NRGD,  ONLY: REGRID_CONTROL     
    USE MESSY_MAIN_GRID_TRAFO_SCRP,  ONLY: t_scrip_data , CALC_SCRIP_WEIGHTS &
                                         , SCRIP_CONTROL, CALC_SCRIPDATA     &
                                         , BALANCE_CURVILINEAR_PS            &
                                         , CONSTRUCT_INPUT_SURF_PRESSURE
    USE MESSY_MAIN_GRID_NETCDF,      ONLY: t_ncvar, COPY_NCVAR               &
                                         , INIT_NCVAR, QDEF_NCVAR            &
                                         , GRD_MAXSTRLEN
    USE MESSY_MAIN_GRID,             ONLY: t_geohybgrid, COPY_GEOHYBGRID     &
                                         , NEW_GEOHYBGRID, LOCATE_GEOHYBGRID &
                                         , INIT_GEOHYBGRID
    USE MESSY_MAIN_IMPORT_GRID_PAR,  ONLY: my_mpi

    IMPLICIT NONE

    INTRINSIC :: PRESENT, SIZE, TRIM

    ! I/O
    ! TODO: use error status in above level
    INTEGER, INTENT (OUT)              :: status       ! error status
    INTEGER, INTENT (IN)               :: iou          ! logical I/O unit
    CHARACTER(LEN=*), INTENT(IN)       :: nmlfile      ! namelist-file
    CHARACTER(LEN=*), INTENT(IN)       :: vname        ! variable name
    INTEGER, INTENT (IN)               :: t            ! netCDF time step
    TYPE (t_ncvar), INTENT(OUT)        :: var          ! nc-variable
    INTEGER, INTENT(IN)                :: iipol        ! interpolation method
    INTEGER, INTENT(IN)                :: ogridid      ! output grid id
    INTEGER, INTENT(INOUT)             :: igridid      ! output grid id
    INTEGER, INTENT(INOUT)             :: SCRIP_ID     ! SCRIP DATA ID   
    LOGICAL, INTENT(IN),  OPTIONAL     :: lrg          ! regrid really ?
    LOGICAL, INTENT(IN),  OPTIONAL     :: lrgx         ! regrid in x
    LOGICAL, INTENT(IN),  OPTIONAL     :: lrgy         ! regrid in y
    LOGICAL, INTENT(IN),  OPTIONAL     :: lrgz         ! regrid in z
    LOGICAL, INTENT(OUT), OPTIONAL     :: lok          ! OK?
    REAL(dp), DIMENSION(:,:), POINTER, OPTIONAL :: oarea
    TYPE(t_geohybgrid),  OPTIONAL, INTENT(INOUT):: convgrid
    LOGICAL, INTENT(IN), OPTIONAL      :: ldompar  ! interpolation runs in domain
                                                   ! parallelisation
    LOGICAL, INTENT(IN), OPTIONAL      :: lvarpar  ! parallise over input vars

    ! LOCAL
    ! OUTPUT OF REGRIDDING PROCEDURE
    CHARACTER(LEN=*), PARAMETER           :: substr = 'RGTOOL_READ_NCVAR'
    LOGICAL,  SAVE                        :: lint    ! use input time ?
! op_pj_20140819+
!!$ INTEGER,  POINTER                     :: RGT(:)        ! regridding type
    INTEGER, DIMENSION(:),  POINTER       :: RGT => NULL() ! regridding type
! op_pj_20140819-
    TYPE (t_ncvar), DIMENSION(:), POINTER :: rvar  => NULL() ! list of variables
    TYPE (t_ncvar), DIMENSION(:), POINTER :: ovar  => NULL() ! list of variables
    TYPE (t_ncvar), DIMENSION(:), POINTER :: sovar => NULL() ! list of variables
    TYPE (t_geohybgrid)                   :: igrid   ! input grid info
    TYPE (t_geohybgrid)                   :: ogrid   ! output grid info
    TYPE (t_geohybgrid)                   :: intgrid ! intermediate grid info
    TYPE (t_geohybgrid)                   :: int2grid! vert. intermed. grid info
    TYPE (t_geohybgrid)                   :: ipsgrid ! intermediate grid info
    TYPE (t_geohybgrid)                   :: intzogrid  ! output grid info
    INTEGER                               :: tt      ! time step
    LOGICAL                               :: llrg    ! regrid really ?
    INTEGER                               :: pcnt    ! procedure counter
    INTEGER                               :: i       ! counter
    TYPE(t_scrip_data), POINTER           :: PSD => NULL() 
    LOGICAL                               :: llrgx, llrgy, llrgz
    INTEGER                               :: ix
    CHARACTER(LEN=4)                      :: cstr = ''
    LOGICAL                               :: lpresax = .FALSE. ! um_ak_20131018 
    LOGICAL                               :: lwork   = .FALSE. ! um_ak_20131024
    
    ! INITIALIZE
    igridid = -99
    status  = 3999

    IF (PRESENT(lok)) THEN
       lok = .FALSE.
    END IF
    IF (PRESENT(lrg)) THEN
       llrg = lrg
    ELSE
       llrg = .true.  ! DEFAULT
    END IF

    IF (PRESENT(lrgx)) THEN
       llrgx = lrgx
    ELSE
       llrgx = .true.  ! DEFAULT
    END IF
    IF (PRESENT(lrgy)) THEN
       llrgy = lrgy
    ELSE
       llrgy = .true.  ! DEFAULT
    END IF
    IF (PRESENT(lrgz)) THEN
       llrgz = lrgz
    ELSE
       llrgz = .false.  ! DEFAULT
    END IF
    !
    NULLIFY(rvar)
    IF (PRESENT(convgrid)) CALL INIT_GEOHYBGRID(convgrid)

    !
    ! LOCATE OUTPUT GRID
    CALL INIT_GEOHYBGRID(ogrid)
    IF (ogridid /= -99) THEN
       CALL LOCATE_GEOHYBGRID(status, ogridid, grid=ogrid)
       IF (status /= 0) RETURN
    END IF

    CALL INIT_NCVAR(var)
    !
    ! USE 1st TIME STEP FOR SCANNING, SINCE A NAMELIST FILE
    ! WITH MORE THAN ONE NAMELIST MUST NOT NECESSARILY
    ! CONTAIN TIME AXES OF EQUAL LENGTH ...
    tt = t
    !
    ! COUNT PROCEDURE LEVEL: 0 : START, VARIABLE NOT YET FOUND IN NAMELIST FILE
    !                        1 : VARIABLE FOUND IN NAMELSIT FILE, BUT NOT YET
    !                            PROCESSED
    !                        2 : VARIABLE FOUND IN NAMELIST FILE, PROCESSING
    !                            ALREADY PERFORMED
    pcnt = 0

    ! START REGRIDDING
    ! REGRIDDING CONTROLLED BY INPUT VARIABELS
    RG_CTRL = RG_SCAN   ! SCAN-MODE, NO REGRIDDING
    RG_NML  = NML_NEXT  ! READ NEXT NAMELIST FROM FILE
    !

    DO ! endless DO loop (must be terminated with EXIT)

       CALL READ_CONTROL(RG_CTRL, RG_NML, RG_STATUS          &
                           , rvar                            &
                           , igrid                           &
                           , ogrid                           &
                           , RGT                             &
                           , lint                            &
                           , TRIM(nmlfile)                   &
                           , tc=tt                           & 
                           , iounit = iou                    &
                           , ldomainpar=ldompar              &
                           , lvarparallel= lvarpar           &
                           , lpresaxis=lpresax               & 
                           , lwork=lwork)                      

       IF (RG_STATUS == RGSTAT_STOP) EXIT

        DO i=1, SIZE(rvar)
          IF (TRIM(rvar(i)%name) == TRIM(vname)) THEN
             ! VARIABLE FOUND IN NAMELIST:
             ! SET PROCEDURE LEVEL
             pcnt = pcnt + 1
             EXIT
          END IF
       END DO

       ! VARIABLE NOT YET FOUND IN NAMELIST FILE:
       ! SCAN NEXT NAMELIST IN NAMELIST FILE
       IF (pcnt == 0) THEN
          RG_NML  = NML_NEXT   ! ... TRY NEXT NAMELIST
          RG_CTRL = RG_SCAN    ! ... SCAN FOR VARIABLE
          
       ELSE
          ! VARIABLE FOUND IN NAMELIST FILE:
          ! SET SWITCHES FOR PROCESSING
          IF (llrg) THEN
             pcnt = pcnt + 1
             RG_CTRL = RG_PROC  ! ... PERFORM REGRIDDING
          ELSE
             RG_CTRL = RG_SCAN  ! ... SCAN AGAIN (= IMPORT RAW DATA)
          END IF
       END IF

       ! FILENAME FOUND IN NAMELIST FILE AND PROCESSING FINISHED:
       ! SET SWITCHES TO END LOOP

       IF (RG_CTRL == RG_PROC) THEN
          SELECT CASE(iipol)
          CASE(GTRF_NONE)
             CALL COPY_NCVAR(var, rvar(i))  ! ... RETURN VARIABLE 
             IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, igrid)
          CASE(GTRF_NRGD)
             IF (lwork) THEN
                CALL REGRID_CONTROL(igrid, ogrid, rvar, ovar, RGT, lint   &
                     , lrgx=llrgx, lrgy=llrgy, lrgz=llrgz, lfirsto=.TRUE. &
                     , lpresaxis=lpresax)
                CALL COPY_NCVAR(var, ovar(i))  ! ... RETURN VARIABLE
             ENDIF
             IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, ogrid)
          CASE(GTRF_SCRP)
             IF (llrgx .OR. llrgy) THEN
                CALL SCRIP_CONTROL(status, SCRIP_ID, igrid, ogrid, RGT, lint &
                     , rvar, sovar, intgrid, llrgz=llrgz, lfirsto=.TRUE.)
                IF (status /= 0) RETURN
             ELSE
                ALLOCATE(sovar(SIZE(rvar)))
                DO ix = 1, SIZE(rvar)
                   CALL COPY_NCVAR(sovar(ix), rvar(ix))
                END DO
                CALL COPY_GEOHYBGRID(intgrid, igrid)
             END IF

             ! TODO vertical interpolation
             IF (llrgz) THEN
                
                ! make grid with vertical input grid, but horizontal
                ! SCRIP-gridded grid 
                CALL COPY_GEOHYBGRID(int2grid, igrid)
                CALL COPY_GEOHYBGRID(intzogrid, ogrid)

                IF ((QDEF_NCVAR(ogrid%rlonm) .AND. QDEF_NCVAR(ogrid%rlatm)) &
                     .OR. &
                     (QDEF_NCVAR(ogrid%rloni) .AND. QDEF_NCVAR(ogrid%rlati)))&
                     THEN

                   ! CONSTRUCT CORRECT PS-INPUT-GRID FOR VERTICAL INTERPOLATION
                   
                   ! SURFACE PRESSURE needs to be available for both grid
                   ! on the rotated grid
                   CALL COPY_GEOHYBGRID(ipsgrid, igrid)
                   ! assume congruency of grids ogrid and rot_ogrid
                   ! take rotated coordinates for curvi-linear grids
                   ! instead of intermediate grid. NCREGRID can only 
                   ! handle lat-lon-grids
                   ipsgrid%lonc = .FALSE.
                   CALL COPY_NCVAR(ipsgrid%lonm, ogrid%rlonm)
                   CALL COPY_NCVAR(ipsgrid%loni, ogrid%rloni)
                   CALL COPY_NCVAR(ipsgrid%latm, ogrid%rlatm)
                   CALL COPY_NCVAR(ipsgrid%lati, ogrid%rlati)
                   
                   CALL INIT_NCVAR(ipsgrid%clonm)
                   CALL INIT_NCVAR(ipsgrid%cloni)
                   CALL INIT_NCVAR(ipsgrid%clatm)
                   CALL INIT_NCVAR(ipsgrid%clati)
                   CALL INIT_NCVAR(ipsgrid%rlonm)
                   CALL INIT_NCVAR(ipsgrid%rloni)
                   CALL INIT_NCVAR(ipsgrid%rlatm)
                   CALL INIT_NCVAR(ipsgrid%rlati)
                   
                   CALL INIT_NCVAR(ipsgrid%ps)
                   CALL CONSTRUCT_INPUT_SURF_PRESSURE(status, igrid, ipsgrid &
                        , SCRIP_ID)
                   IF (status /= 0) RETURN

                   CALL BALANCE_CURVILINEAR_PS(status, ipsgrid, ogrid &
                        , SCRIP_ID)
                   IF (status /= 0) RETURN
                   ! replace in int2grid for vertical interpolation
                   CALL COPY_GEOHYBGRID(int2grid, ipsgrid)
                   CALL INIT_GEOHYBGRID(ipsgrid)

                   ! CONSTRUCT OUTPUT GRID FOR VERTICAL INTERPOLATION
                   intzogrid%lonc = .FALSE.
                   CALL COPY_NCVAR(intzogrid%lonm, ogrid%rlonm)
                   CALL COPY_NCVAR(intzogrid%loni, ogrid%rloni)
                   CALL COPY_NCVAR(intzogrid%latm, ogrid%rlatm)
                   CALL COPY_NCVAR(intzogrid%lati, ogrid%rlati)
                   
                   CALL INIT_NCVAR(intzogrid%clonm)
                   CALL INIT_NCVAR(intzogrid%cloni)
                   CALL INIT_NCVAR(intzogrid%clatm)
                   CALL INIT_NCVAR(intzogrid%clati)
                   CALL INIT_NCVAR(intzogrid%rlonm)
                   CALL INIT_NCVAR(intzogrid%rloni)
                   CALL INIT_NCVAR(intzogrid%rlatm)
                   CALL INIT_NCVAR(intzogrid%rlati)
                ELSE
                   ! no curvi-linear grid, used horizontal grid from
                   ! intermediate grid
                   CALL COPY_NCVAR(int2grid%lonm, intgrid%lonm)
                   CALL COPY_NCVAR(int2grid%loni, intgrid%loni)
                   CALL COPY_NCVAR(int2grid%latm, intgrid%latm)
                   CALL COPY_NCVAR(int2grid%lati, intgrid%lati)
                   CALL CONSTRUCT_INPUT_SURF_PRESSURE(status, igrid &
                        , int2grid, SCRIP_ID)
                   IF (status /= 0) RETURN
                   CALL BALANCE_CURVILINEAR_PS(status, int2grid, ogrid &
                        , SCRIP_ID)
                   IF (status /= 0) RETURN
                ENDIF
                
                IF (lwork) THEN
                   CALL REGRID_CONTROL(int2grid, intzogrid, sovar, ovar, RGT &
                        , lint , lrgx=.FALSE., lrgy=.FALSE., lrgz=llrgz      &
                        , lfirsto=.TRUE., lpresaxis=lpresax)
                   CALL COPY_NCVAR(var, ovar(i))
                ENDIF          
                
                IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, ogrid)
                CALL INIT_GEOHYBGRID(int2grid)
                CALL INIT_GEOHYBGRID(intzogrid)
             ELSE
                CALL COPY_NCVAR(var, sovar(i))
                IF (PRESENT(convgrid)) &
                     CALL COPY_GEOHYBGRID(convgrid, intgrid)
             ENDIF

             CALL INIT_GEOHYBGRID(intgrid)
          END SELECT
       ELSE
          CALL COPY_NCVAR(var, rvar(i))  ! ... RETURN VARIABLE
! um_ak_20150714+
! names are now generated in new_geohybgrid
!!$          COUNT = COUNT + 1
!!$          CALL int2str(cstr, COUNT, '0')
!!$          igrid%name = 'ingrid'//cstr
! um_ak_20150714-
          CALL NEW_GEOHYBGRID(status, igridid, igrid) 
          IF (status /= 0 .AND. status /= 1) RETURN
          
          IF (iipol == GTRF_SCRP) THEN
             CALL CALC_SCRIPDATA(status, igrid, ogrid, RGT, SCRIP_ID &
                  , oarea, PSD)
             IF (status /= 0 ) THEN
                IF (status /= 1 )  RETURN
             ELSE
                ! CALCULATE WEIGHTS
                CALL CALC_SCRIP_WEIGHTS(status, PSD)
             END IF
             NULLIFY(PSD) ! um_ak_20140313
          ENDIF
          IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, igrid)
       END IF
       RG_CTRL = RG_STOP              ! ... TERMINATE ...
       ! ... REGRIDDING PROCEDURE CORRECTLY
       IF (PRESENT(lok)) THEN
          lok = .TRUE.                        ! ... REPORT SUCCESS
       END IF
       
       IF (ASSOCIATED(rvar)) THEN
          DO i=1, SIZE(rvar)
             CALL INIT_NCVAR(rvar(i))
          END DO
          DEALLOCATE(rvar)
          NULLIFY(rvar)
       END IF

       IF (ASSOCIATED(ovar)) THEN
          DO i=1, SIZE(ovar)
             CALL INIT_NCVAR(ovar(i))
          END DO
          DEALLOCATE(ovar)
          NULLIFY(ovar) 
       ENDIF

      IF (ASSOCIATED(sovar)) THEN
          DO i=1, SIZE(sovar)
             CALL INIT_NCVAR(sovar(i))
          END DO
          DEALLOCATE(sovar)
          NULLIFY(sovar) 
       ENDIF

       IF  (ASSOCIATED(RGT)) DEALLOCATE(RGT) ! um_ak_20140314
    END DO  ! ENDLESS DO-LOOP ...

    ! op_jp_20140222+
    ! FREE MEMORY
    CALL INIT_GEOHYBGRID(ogrid)
    CALL INIT_GEOHYBGRID(igrid)
    ! op_pj_20140222-
    
    ! END REGRIDDING
    status = 0

  END SUBROUTINE RGTOOL_READ_NCVAR
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE RGTOOL_READ_NCFILE(status, iou, nmlfile, fname, t, var &
                               , iipol, ogridid, igridid, SCRIP_ID  &
                               , lrg, lrgx, lrgy, lrgz, lok, oarea  &
                               , convgrid, ldompar, lvarpar)

    ! PERFORMS ONE RE-GRIDDING (TIME-)STEP (t) FOR ONE FILE
    ! (fname) AND RETURNS LIST OF TYPE NCVAR (var)
    ! AND OPTIONALLY THE GRID (grid)
    ! LOGICAL I/O UNIT  : iou          (for namelist-file)
    ! NAMELIST-FILE     : nmlfile
    ! DIMENSION-SWITCHES: lrgx, lrgy, lrgz
    ! REGRIDDING-SWITCH : lrg
    ! SUCCESS-FLAG      : lok
    !
    ! IF fname = '' THEN PROCESS 1st NAMELIST ONLY !!!
    !
    ! Author: Patrick Joeckel, MPICH, October 2002
    !         Astrid Kerkweg,  UniMz, 2013 (restrctured + expanded)
    
    USE MESSY_MAIN_TOOLS,             ONLY: INT2STR
    ! NCREGRID INTERFACE
    USE MESSY_MAIN_GRID_TRAFO,        ONLY: GTRF_NONE, GTRF_NRGD, GTRF_SCRP
    USE MESSY_MAIN_GRID_TRAFO_NRGD,   ONLY: REGRID_CONTROL
    USE MESSY_MAIN_GRID_TRAFO_SCRP,   ONLY: t_scrip_data , CALC_SCRIP_WEIGHTS &
                                          , SCRIP_CONTROL, CALC_SCRIPDATA     &
                                          , BALANCE_CURVILINEAR_PS            &
                                          , CONSTRUCT_INPUT_SURF_PRESSURE
    USE MESSY_MAIN_GRID_NETCDF,       ONLY: t_ncvar, COPY_NCVAR, ERRMSG       &
                                          , QDEF_NCVAR, INIT_NCVAR            &
                                          , GRD_MAXSTRLEN
    USE MESSY_MAIN_GRID,              ONLY: t_geohybgrid,   COPY_GEOHYBGRID   &
                                          , NEW_GEOHYBGRID, LOCATE_GEOHYBGRID &
                                          , INIT_GEOHYBGRID
    USE MESSY_MAIN_IMPORT_GRID_PAR,   ONLY: my_mpi

    IMPLICIT NONE

    INTRINSIC :: PRESENT, SIZE, TRIM

    ! I/O
    ! TODO: use error status in above level
    INTEGER, INTENT (OUT)                 :: status       ! error status
    INTEGER, INTENT (IN)                  :: iou          ! logical I/O unit
    CHARACTER(LEN=*), INTENT(IN)          :: nmlfile      ! namelist-file
    CHARACTER(LEN=*), INTENT(IN)          :: fname        ! filename
    INTEGER, INTENT (IN)                  :: t            ! netCDF time step
    TYPE (t_ncvar), DIMENSION(:), POINTER :: var          ! nc-variables
    INTEGER, INTENT(IN)                   :: iipol        ! interpolation method
    INTEGER, INTENT(INOUT)                :: igridid
    INTEGER, INTENT(IN)                   :: ogridid
    INTEGER, INTENT(INOUT)                :: SCRIP_ID     ! SCRIP DATA ID   
    LOGICAL, INTENT(IN),   OPTIONAL       :: lrg          ! regrid really ?
    LOGICAL, INTENT(IN),   OPTIONAL       :: lrgx         ! regrid in x
    LOGICAL, INTENT(IN),   OPTIONAL       :: lrgy         ! regrid in y
    LOGICAL, INTENT(IN),   OPTIONAL       :: lrgz         ! regrid in z
    LOGICAL, INTENT (OUT), OPTIONAL       :: lok          ! OK?
    TYPE(t_geohybgrid), INTENT(INOUT), OPTIONAL :: convgrid
    REAL(dp), DIMENSION(:,:), POINTER, OPTIONAL :: oarea
    LOGICAL, INTENT(IN),   OPTIONAL       :: ldompar  ! interpolation runs in 
                                                      ! domain parallelisation
    LOGICAL, INTENT(IN),   OPTIONAL       :: lvarpar  ! parallise over input vars

    ! LOCAL
    ! OUTPUT OF REGRIDDING PROCEDURE
    CHARACTER(LEN=*), PARAMETER           :: substr = 'RGTOOL_READ_NCFILE'
    LOGICAL,  SAVE                        :: lint    ! use input time ?
! op_pj_20140819+
!!$ INTEGER,  POINTER                     :: RGT(:)        ! regridding type
    INTEGER, DIMENSION(:),  POINTER       :: RGT => NULL() ! regridding type
! op_pj_20140819-
    TYPE (t_ncvar), DIMENSION(:), POINTER :: rvar    ! list of variables
    TYPE (t_ncvar), DIMENSION(:), POINTER :: ovar    ! list of variables
    TYPE (t_ncvar), DIMENSION(:), POINTER :: sovar   ! list of variables
    TYPE (t_geohybgrid)                   :: igrid   ! output grid info
    TYPE (t_geohybgrid)                   :: ogrid   ! output grid info
    TYPE (t_geohybgrid)                   :: intgrid ! intermediate grid info
    TYPE (t_geohybgrid)                   :: int2grid! intermediate grid info
    TYPE (t_geohybgrid)                   :: ipsgrid ! intermediate grid info
    TYPE (t_geohybgrid)                   :: intzogrid  ! output grid info
    INTEGER                               :: tt      ! time step
    LOGICAL                               :: llrg    ! regrid really ?
    INTEGER                               :: pcnt    ! procedure counter
    INTEGER                               :: i       ! counter
    TYPE(t_scrip_data), POINTER           :: PSD
    LOGICAL                               :: llrgx, llrgy, llrgz
    INTEGER                               :: ix
    CHARACTER(LEN=4)                      :: cstr = ''
    LOGICAL                               :: lpresax = .FALSE. 
    LOGICAL                               :: lwork   = .FALSE. 

    ! INITIALIZE
    igridid = -99
    status  = 3999

    IF (PRESENT(lok)) THEN
       lok = .FALSE.
    END IF
    IF (PRESENT(lrg)) THEN
       llrg = lrg
    ELSE
       llrg = .true.  ! DEFAULT
    END IF
    IF (PRESENT(lrgx)) THEN
       llrgx = lrgx
    ELSE
       llrgx = .true.  ! DEFAULT
    END IF
    IF (PRESENT(lrgy)) THEN
       llrgy = lrgy
    ELSE
       llrgy = .true.  ! DEFAULT
    END IF
    IF (PRESENT(lrgz)) THEN
       llrgz = lrgz
    ELSE
       llrgz = .false.  ! DEFAULT
    END IF
    !
    NULLIFY(rvar)
    NULLIFY(ovar)
    NULLIFY(sovar)
    !
    IF (PRESENT(convgrid)) CALL INIT_GEOHYBGRID(convgrid)

    ! LOCATE OUTPUT GRID
    CALL INIT_GEOHYBGRID(ogrid)
    IF (ogridid /= -99) THEN
       CALL LOCATE_GEOHYBGRID(status, ogridid, grid=ogrid)
       IF (status /= 0) RETURN
    END IF

    !
    ! USE 1st TIME STEP FOR SCANNING, SINCE A NAMELIST FILE
    ! WITH MORE THAN ONE NAMELIST MUST NOT NECESSARILY
    ! CONTAIN TIME AXES OF EQUAL LENGTH ...
    tt = t
    !
    ! COUNT PROCEDURE LEVEL: 0 : START, FILENAME NOT YET FOUND IN NAMELIST FILE
    !                        1 : FILENAME FOUND IN NAMELSIT FILE, BUT NOT YET
    !                            PROCESSED
    !                        2 : FILENAME FOUND IN NAMELIST FILE, PROCESSING
    !                            ALREADY PERFORMED
    pcnt = 0

    ! START REGRIDDING
    ! REGRIDDING CONTROLLED BY INPUT VARIABELS
    RG_CTRL = RG_SCAN   ! SCAN-MODE, NO REGRIDDING
    RG_NML  = NML_NEXT  ! READ NEXT NAMELIST FROM FILE
    !
    DO ! endless DO loop (must be terminated with EXIT)

       CALL READ_CONTROL(RG_CTRL, RG_NML, RG_STATUS        &
                           , rvar                          &
                           , igrid                         &
                           , ogrid                         &
                           , RGT                           &
                           , lint                          &
                           , TRIM(nmlfile)                 &
                           , tc=tt                         & 
                           , iounit = iou                  &
                           , ldomainpar=ldompar            &
                           , lvarparallel= lvarpar         &
                           , lpresaxis=lpresax             & 
                           , lwork=lwork)

       IF (RG_STATUS == RGSTAT_STOP) EXIT

       ! FILENAME FOUND IN NAMELIST OR ALREADY PROCESSED
       IF ((TRIM(ogrid%file) == TRIM(fname)).OR.(pcnt == 1) &
            .OR. (TRIM(fname) == '') ) THEN
          ! SET PROCEDURE LEVEL
          pcnt = pcnt + 1
       END IF

       ! FILENAME NOT YET FOUND IN NAMELIST FILE:
       ! SCAN NEXT NAMELIST IN NAMELIST FILE
       IF (pcnt == 0) THEN
          RG_NML  = NML_NEXT   ! ... TRY NEXT NAMELIST
          RG_CTRL = RG_SCAN    ! ... SCAN FOR VARIABLE

       ELSE
       ! FILENAME FOUND IN NAMELIST FILE:
       ! SET SWITCHES FOR PROCESSING
          IF (llrg) THEN
             RG_CTRL = RG_PROC  ! ... PERFORM REGRIDDING
          ELSE
             RG_CTRL = RG_SCAN  ! ... COPY RAW DATA (= IMPORT RAW DATA)
          END IF
       END IF

       ! FILENAME FOUND IN NAMELIST FILE AND PROCESSING FINISHED:
       ! SET SWITCHES TO END LOOP
       IF (RG_CTRL == RG_PROC) THEN
          SELECT CASE(iipol)
          CASE(GTRF_NONE)
             ALLOCATE(var(SIZE(rvar)), STAT=status)
             CALL ERRMSG(substr,status,1)
             DO i=1, SIZE(rvar)
                CALL COPY_NCVAR(var(i), rvar(i))    ! ... RETURN VARIABLES
             END DO
             IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, igrid)
          CASE(GTRF_NRGD)
             IF (lwork) THEN
                CALL REGRID_CONTROL(igrid, ogrid, rvar, ovar, RGT, lint  &
                     , lrgx=llrgx, lrgy=llrgy, lrgz=llrgz, lfirsto=.TRUE.&
                     , lpresaxis=lpresax)
                ALLOCATE(var(SIZE(ovar)), STAT=status)
                CALL ERRMSG(substr,status,1)
                DO i=1, SIZE(ovar)
                   CALL COPY_NCVAR(var(i), ovar(i))    ! ... RETURN VARIABLES
                END DO
             ENDIF
             IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, ogrid)
          CASE(GTRF_SCRP)
             IF (llrgx .OR. llrgy) THEN
                CALL SCRIP_CONTROL(status, SCRIP_ID, igrid, ogrid, RGT, lint&
                     , rvar, sovar,intgrid, llrgz=llrgz, lfirsto=.TRUE.)
                IF (status /= 0) RETURN
             ELSE
                ALLOCATE(sovar(SIZE(rvar)))
                DO ix = 1, SIZE(rvar)
                   CALL COPY_NCVAR(sovar(ix), rvar(ix))
                END DO
                CALL COPY_GEOHYBGRID(intgrid, igrid)
             END IF

             ! vertical interpolation
             IF (llrgz) THEN
                   
                ! make grid with vertical input grid, but horizontal
                ! SCRIP-gridded grid 
                CALL COPY_GEOHYBGRID(int2grid,  igrid)
                CALL COPY_GEOHYBGRID(intzogrid, ogrid)
                
                IF ((QDEF_NCVAR(ogrid%rlonm) .AND. QDEF_NCVAR(ogrid%rlatm)) &
                     .OR. &
                     (QDEF_NCVAR(ogrid%rloni) .AND. QDEF_NCVAR(ogrid%rlati)))&
                     THEN
                   
                   ! CONSTRUCT CORRECT PS-INPUT-GRID FOR VERTICAL INTERPOLATION

                   ! SURFACE PRESSURE needs to be available for both grid
                   ! on the rotated grid
                   CALL COPY_GEOHYBGRID(ipsgrid, igrid)
                   
                   ! assume congruency of grids ogrid and intgrid
                   ! take rotated coordinates for curvi-linear grids
                   ! instead of intermediate grid. NCREGRID can only 
                   ! handle lat-lon-grids
                   ipsgrid%lonc = .FALSE.
                   CALL COPY_NCVAR(ipsgrid%lonm, ogrid%rlonm)
                   CALL COPY_NCVAR(ipsgrid%loni, ogrid%rloni)
                   CALL COPY_NCVAR(ipsgrid%latm, ogrid%rlatm)
                   CALL COPY_NCVAR(ipsgrid%lati, ogrid%rlati)
                   
                   intzogrid%lonc = .FALSE.
                   CALL COPY_NCVAR(intzogrid%lonm, ogrid%rlonm)
                   CALL COPY_NCVAR(intzogrid%loni, ogrid%rloni)
                   CALL COPY_NCVAR(intzogrid%latm, ogrid%rlatm)
                   CALL COPY_NCVAR(intzogrid%lati, ogrid%rlati)

                   CALL INIT_NCVAR(int2grid%clonm)
                   CALL INIT_NCVAR(int2grid%cloni)
                   CALL INIT_NCVAR(int2grid%clatm)
                   CALL INIT_NCVAR(int2grid%clati)
                   CALL INIT_NCVAR(int2grid%rlonm)
                   CALL INIT_NCVAR(int2grid%rloni)
                   CALL INIT_NCVAR(int2grid%rlatm)
                   CALL INIT_NCVAR(int2grid%rlati)
                      
                   ! construct ps
                   CALL INIT_NCVAR(ipsgrid%ps)
                   CALL CONSTRUCT_INPUT_SURF_PRESSURE(status, igrid, ipsgrid &
                                                          , SCRIP_ID)
                   IF (status /= 0) RETURN
                   CALL BALANCE_CURVILINEAR_PS(status, ipsgrid, ogrid &
                        , SCRIP_ID)
                   IF (status /= 0) RETURN
                   ! replace in int2grid for vertical interpolation
                   CALL COPY_GEOHYBGRID(int2grid, ipsgrid)
                   CALL INIT_GEOHYBGRID(ipsgrid)
                      
                   ! CONSTRUCT OUTPUT GRID FOR VERTICAL INTERPOLATION
                   CALL INIT_NCVAR(intzogrid%clonm)
                   CALL INIT_NCVAR(intzogrid%cloni)
                   CALL INIT_NCVAR(intzogrid%clatm)
                   CALL INIT_NCVAR(intzogrid%clati)
                   CALL INIT_NCVAR(intzogrid%rlonm)
                   CALL INIT_NCVAR(intzogrid%rloni)
                   CALL INIT_NCVAR(intzogrid%rlatm)
                   CALL INIT_NCVAR(intzogrid%rlati)
                ELSE
                   ! no curvi-linear grid, used horizontal grid from
                   ! intermediate grid
                   CALL COPY_NCVAR(int2grid%lonm, intgrid%lonm)
                   CALL COPY_NCVAR(int2grid%loni, intgrid%loni)
                   CALL COPY_NCVAR(int2grid%latm, intgrid%latm)
                   CALL COPY_NCVAR(int2grid%lati, intgrid%lati)
                   
                   CALL CONSTRUCT_INPUT_SURF_PRESSURE(status, igrid &
                        , int2grid, SCRIP_ID)
                   IF (status /= 0) RETURN
                   CALL BALANCE_CURVILINEAR_PS(status, int2grid, ogrid &
                        , SCRIP_ID)
                   IF (status /= 0) RETURN
                ENDIF
                
                IF (lwork) THEN
                   CALL REGRID_CONTROL(int2grid, intzogrid, sovar, ovar, RGT &
                        , lint, lrgx=.FALSE., lrgy=.FALSE., lrgz=llrgz       &
                        , lfirsto=.TRUE., lpresaxis=lpresax)

                   ALLOCATE(var(SIZE(ovar)), STAT=status)
                   CALL ERRMSG(substr,status,1)
                   DO i=1, SIZE(ovar)
                      CALL COPY_NCVAR(var(i), ovar(i))  ! ... RETURN VARIABLES
                   END DO
                END IF
                IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, ogrid)
                
                CALL INIT_GEOHYBGRID(int2grid)
                CALL INIT_GEOHYBGRID(intzogrid)
             ELSE
                ALLOCATE(var(SIZE(sovar)), STAT=status)
                CALL ERRMSG(substr,status,1)
                DO i=1, SIZE(sovar)
                   CALL COPY_NCVAR(var(i), sovar(i))   ! ... RETURN VARIABLES
                END DO
                IF (PRESENT(convgrid)) &
                     CALL COPY_GEOHYBGRID(convgrid, intgrid)
                
             ENDIF
             CALL INIT_GEOHYBGRID(intgrid)
          END SELECT
       ELSE
          ALLOCATE(var(SIZE(rvar)), STAT=status)
          CALL ERRMSG(substr,status,1)
          DO i=1, SIZE(rvar)
             CALL COPY_NCVAR(var(i), rvar(i))    ! ... RETURN VARIABLES
          END DO
! um_ak_201507014+
! grid names are now generated in NEW_GEOHYBGRID
!!$          COUNT = COUNT + 1
!!$          CALL int2str(cstr, COUNT, '0')
!!$          igrid%name = 'ingrid'//cstr
! um_ak_201507014-
          CALL NEW_GEOHYBGRID(status, igridid, igrid) 
          IF (status /= 0 .AND. status /= 1) RETURN 
          IF (iipol == GTRF_SCRP) THEN
             CALL CALC_SCRIPDATA(status, igrid, ogrid, RGT, SCRIP_ID &
                  , oarea, PSD)
             IF (status /= 0 ) THEN
                ! CALCULATE WEIGHTS
                IF (status /= 01 )  RETURN
             ELSE
                CALL CALC_SCRIP_WEIGHTS(status, PSD)
             END IF
             NULLIFY(PSD) ! um_ak_20140313
          ENDIF
          IF (PRESENT(convgrid)) CALL COPY_GEOHYBGRID(convgrid, igrid)
          
       END IF
       RG_CTRL = RG_STOP              ! ... TERMINATE ...
                                         ! ... REGRIDDING PROCEDURE CORRECTLY
       IF (PRESENT(lok)) THEN
          lok = .TRUE.                        ! ... REPORT SUCCESS
       END IF

       IF (ASSOCIATED(rvar)) THEN
          DO i=1, SIZE(rvar)
             CALL INIT_NCVAR(rvar(i))
          END DO
          DEALLOCATE(rvar)
          NULLIFY(rvar)
       END IF

       IF (ASSOCIATED(ovar)) THEN
          DO i=1, SIZE(ovar)
             CALL INIT_NCVAR(ovar(i))
          END DO
          DEALLOCATE(ovar)
          NULLIFY(ovar) 
       ENDIF

      IF (ASSOCIATED(sovar)) THEN
          DO i=1, SIZE(sovar)
             CALL INIT_NCVAR(sovar(i))
          END DO
          DEALLOCATE(sovar)
          NULLIFY(sovar) 
       ENDIF

       IF  (ASSOCIATED(RGT)) THEN
          DEALLOCATE(RGT) ! um_ak_20140314
          NULLIFY(RGT)    ! op_pj_20140820
       END IF

    END DO  ! ENDLESS DO-LOOP ...
    ! END REGRIDDING

! op_pj_20140222+
    ! FREE MEMORY
    CALL INIT_GEOHYBGRID(ogrid)
    CALL INIT_GEOHYBGRID(igrid)
! op_pj_20140222-

    status = 0

  END SUBROUTINE RGTOOL_READ_NCFILE

END MODULE MESSY_MAIN_IMPORT_GRID
! ------------------------------------------------------------------
! ******************************************************************
